"""Evaluator configuration classes.

This module defines the configuration hierarchy for evaluators in the new veeksha
framework. Evaluators are responsible for computing metrics from benchmark runs.

The hierarchy follows the BasePolyConfig pattern used elsewhere in veeksha:
- BaseEvaluatorConfig (abstract base)
  - PerformanceEvaluatorConfig (latency, throughput)
  - LMEvalAccuracyEvaluatorConfig (task-specific correctness - lm-eval)
"""

from dataclasses import field
from typing import Optional, Union

from veeksha.config.core.base_poly_config import BasePolyConfig
from veeksha.config.core.frozen_dataclass import frozen_dataclass
from veeksha.config.slo import BaseSloConfig, ConstantSloConfig
from veeksha.types import ChannelModality, EvaluationType, SloType


@frozen_dataclass
class DecodeWindowConfig:
    """Configuration for decode window analysis.

    Filters metrics to the window where the server is doing decode
    with a full batch of requests.
    """

    min_active_requests: Union[int, str] = field(
        default=1,
        metadata={
            "help": "Minimum number of simultaneously generating (decoding) requests "
            "required for a time interval to be considered inside the decode window. "
            "Use 'max_observed' to auto-detect the peak concurrent decoding count."
        },
    )
    selection_strategy: str = field(
        default="longest",
        metadata={
            "help": "Which window(s) to analyze when multiple windows exist. "
            "Supported: 'longest' (single longest), 'first' (single first), "
            "'all' (aggregate all qualifying windows)."
        },
    )
    anchor_to_client_pickup: bool = field(
        default=True,
        metadata={
            "help": "If True, anchor per-request token times to client_picked_up_at "
            "when available; otherwise use scheduler_dispatched_at."
        },
    )
    require_streaming: bool = field(
        default=True,
        metadata={
            "help": "If True, only streaming requests contribute to decode window analysis."
        },
    )

    def __post_init__(self):
        if isinstance(self.min_active_requests, int):
            if self.min_active_requests <= 0:
                raise ValueError("min_active_requests must be > 0")
        elif isinstance(self.min_active_requests, str):
            if self.min_active_requests != "max_observed":
                raise ValueError(
                    f"Invalid min_active_requests '{self.min_active_requests}'. "
                    "Supported string value: 'max_observed'"
                )
        else:
            raise ValueError("min_active_requests must be int or 'max_observed'")
        allowed = {"longest", "first", "all"}
        if self.selection_strategy not in allowed:
            raise ValueError(
                f"Invalid selection_strategy '{self.selection_strategy}'. "
                f"Supported: {sorted(allowed)}"
            )


# ---- Channel-specific performance configs ----


@frozen_dataclass
class BaseChannelPerformanceConfig(BasePolyConfig):
    """Base config for channel-specific performance"""


@frozen_dataclass
class TextChannelPerformanceConfig(BaseChannelPerformanceConfig):
    """Text channel performance configuration"""

    decode_window_enabled: bool = field(
        default=False, metadata={"help": "Enable decode window analysis"}
    )
    decode_window_config: Optional[DecodeWindowConfig] = field(
        default=None,
        metadata={"help": "Decode window configuration (required if enabled)"},
    )

    @classmethod
    def get_type(cls) -> ChannelModality:
        return ChannelModality.TEXT

    def __post_init__(self):
        if self.decode_window_enabled and self.decode_window_config is None:
            raise ValueError(
                "decode_window_config is required when decode_window_enabled=True"
            )


class ImageChannelPerformanceConfig(BaseChannelPerformanceConfig):
    """Image channel performance configuration"""

    @classmethod
    def get_type(cls) -> ChannelModality:
        return ChannelModality.IMAGE


@frozen_dataclass
class AudioChannelPerformanceConfig(BaseChannelPerformanceConfig):
    """Audio channel performance configuration"""

    @classmethod
    def get_type(cls) -> ChannelModality:
        return ChannelModality.AUDIO


@frozen_dataclass
class VideoChannelPerformanceConfig(BaseChannelPerformanceConfig):
    """Video channel performance configuration"""

    @classmethod
    def get_type(cls) -> ChannelModality:
        return ChannelModality.VIDEO


# ---- Base evaluator config ----


def _default_slos() -> list[BaseSloConfig]:
    return [
        ConstantSloConfig(
            metric="ttfc",
            percentile=0.99,
            value=0.5,
            name="P99 TTFC",
        ),
        ConstantSloConfig(
            metric="tbc",
            percentile=0.9,
            value=0.05,
            name="P90 TBC",
        ),
    ]


@frozen_dataclass(allow_from_file=True)
class BaseEvaluatorConfig(BasePolyConfig):
    """Base configuration for all evaluators (performance, accuracy)"""

    target_channels: list = field(
        default_factory=lambda: ["text"],
        metadata={"help": "List of ChannelModality values to evaluate."},
    )

    slos: list[BaseSloConfig] = field(
        default_factory=_default_slos,
        metadata={
            "help": f"List of SLO definitions to evaluate against request-level metrics. {SloType.help_str()}"
        },
    )

    stream_metrics: bool = field(
        default=True, metadata={"help": "Enable real-time metric streaming"}
    )
    stream_metrics_interval: float = field(
        default=5.0, metadata={"help": "Interval for streaming metrics in seconds"}
    )

    def __post_init__(self):
        if self.target_channels:
            converted = []
            for ch in self.target_channels:
                if isinstance(ch, str):
                    converted.append(ChannelModality.from_str(ch))
                else:
                    converted.append(ch)
            object.__setattr__(self, "target_channels", converted)


# ---- Performance evaluator config ----


@frozen_dataclass
class PerformanceEvaluatorConfig(BaseEvaluatorConfig):
    """Configuration for system performance evaluation."""

    text_channel: TextChannelPerformanceConfig = field(
        default_factory=TextChannelPerformanceConfig,
        metadata={"help": "Text channel performance configuration"},
    )
    image_channel: ImageChannelPerformanceConfig = field(
        default_factory=ImageChannelPerformanceConfig,
        metadata={"help": "Image channel performance configuration"},
    )
    audio_channel: Optional[AudioChannelPerformanceConfig] = field(
        default=None,
        metadata={"help": "Audio channel performance configuration"},
    )
    video_channel: Optional[VideoChannelPerformanceConfig] = field(
        default=None,
        metadata={"help": "Video channel performance configuration"},
    )

    @classmethod
    def get_type(cls) -> EvaluationType:
        return EvaluationType.PERFORMANCE

    def get_channel_config(
        self, channel: ChannelModality
    ) -> Optional[BaseChannelPerformanceConfig]:
        """Get the performance config for a specific channel."""
        if channel == ChannelModality.TEXT:
            return self.text_channel
        elif channel == ChannelModality.IMAGE:
            return self.image_channel
        elif channel == ChannelModality.AUDIO:
            return self.audio_channel
        elif channel == ChannelModality.VIDEO:
            return self.video_channel
        return None


# ---- Accuracy evaluator config(s) ----


@frozen_dataclass
class LMEvalAccuracyEvaluatorConfig(BaseEvaluatorConfig):
    """Configuration for lm-eval accuracy evaluation (task-specific correctness).

    IMPORTANT: For lm-eval accuracy evaluation, the content generation must use
    `LMEvalSessionGenerator`. The generator owns the lm-eval Task/Instance objects,
    and the evaluator binds responses to instances for evaluation.
    """

    bootstrap_iters: int = field(
        default=100000,
        metadata={"help": "Bootstrap iterations for confidence intervals"},
    )

    @classmethod
    def get_type(cls) -> EvaluationType:
        return EvaluationType.ACCURACY_LMEVAL



from dataclasses import dataclass, field
from typing import Dict, List, Optional


@dataclass
class SessionNode:
    id: int
    wait_after_ready: float


@dataclass
class SessionEdge:
    src: int
    dst: int
    is_history_parent: bool = True


@dataclass
class SessionGraph:
    nodes: Dict[int, SessionNode] = field(default_factory=dict)
    outgoing: Dict[int, List[SessionEdge]] = field(
        default_factory=dict
    )  # outgoing edges from a node
    incoming: Dict[int, List[SessionEdge]] = field(
        default_factory=dict
    )  # incoming edges to a node


def add_node(graph: SessionGraph, node: SessionNode) -> None:
    if node.id in graph.nodes:
        raise ValueError(f"Node {node.id} already exists")
    graph.nodes[node.id] = node


def add_edge(graph: SessionGraph, edge: SessionEdge) -> None:
    if edge.src not in graph.nodes or edge.dst not in graph.nodes:
        raise ValueError("Both endpoints must exist before adding an edge")
    graph.outgoing.setdefault(edge.src, []).append(edge)
    graph.incoming.setdefault(edge.dst, []).append(edge)


def parents(graph: SessionGraph, node_id: int) -> List[SessionEdge]:
    return graph.incoming.get(node_id, [])


def children(graph: SessionGraph, node_id: int) -> List[SessionEdge]:
    return graph.outgoing.get(node_id, [])


def get_node_ids(graph: SessionGraph) -> List[int]:
    return list(graph.nodes.keys())


def is_root(graph: SessionGraph, node_id: int) -> bool:
    return len(parents(graph, node_id)) == 0


def ready_at(
    graph: SessionGraph, node_id: int, completions: Dict[int, float]
) -> Optional[float]:
    """
    Returns the time at which the node is ready to be executed.

    If a parent completion time is missing, return None
    """
    ps = parents(graph, node_id)
    if not ps:
        return graph.nodes[node_id].wait_after_ready
    parent_times = []
    for edge in ps:
        if edge.src not in completions:
            return None
        parent_times.append(completions[edge.src])
    parent_finish = max(parent_times)
    return parent_finish + graph.nodes[node_id].wait_after_ready


def is_ready(
    graph: SessionGraph, node_id: int, completions: Dict[int, float], now: float
) -> bool:
    ready_time = ready_at(graph, node_id, completions)
    if ready_time is None:
        return False
    return now >= ready_time


def topological_order(graph: SessionGraph) -> List[int]:
    incoming_counts = {nid: len(parents(graph, nid)) for nid in graph.nodes}
    queue = [nid for nid, deg in incoming_counts.items() if deg == 0]
    order: List[int] = []
    while queue:
        current = queue.pop()
        order.append(current)
        for edge in children(graph, current):
            incoming_counts[edge.dst] -= 1
            if incoming_counts[edge.dst] == 0:
                queue.append(edge.dst)
    if len(order) != len(graph.nodes):
        raise ValueError("SessionGraph contains a cycle")
    return order


def format_session_graph(graph: SessionGraph) -> str:
    lines = []
    lines.append("  Nodes:")
    for node_id, node in sorted(graph.nodes.items()):
        lines.append(f"    {node_id} -> wait_after_ready={node.wait_after_ready}")
    lines.append("  Edges:")
    seen_edges = []
    for edges in graph.outgoing.values():
        seen_edges.extend(edges)
    for edge in sorted(seen_edges, key=lambda e: (e.src, e.dst)):
        lines.append(f"    {edge.src} -> {edge.dst}")
    return "\n".join(lines)


def print_session_graph(graph: SessionGraph) -> None:
    print(format_session_graph(graph))


def render_session_graph(
    graph: SessionGraph,
    output_path: Optional[str] = None,
    format: str = "png",
    view: bool = False,
) -> str:
    """
    Render session graph using Graphviz.

    Generates a proper graph visualization with:
    - Nodes arranged by depth (layers)
    - Edges with arrows and proper routing
    - History parent edges in bold/red
    - Skip connections visible

    Args:
        graph: The SessionGraph to render
        output_path: Path to save the image (without extension).
                     If None, returns DOT source only.
        format: Output format (png, svg, pdf)
        view: If True, open the rendered image

    Returns:
        DOT source string

        >>> from veeksha.core.session_graph import render_session_graph
        >>> render_session_graph(graph, "my_graph", format="png")
        # Creates my_graph.png
    """
    try:
        import graphviz  # type: ignore
    except ImportError:
        return (
            "ERROR: graphviz not installed. Install with:\n"
            "  pip install graphviz\n"
            "And ensure Graphviz is installed on your system:\n"
            "  apt-get install graphviz  # Ubuntu/Debian\n"
            "  brew install graphviz     # macOS"
        )

    # Create directed graph
    dot = graphviz.Digraph(comment="Session Graph")
    dot.attr(rankdir="TB")  # Top to bottom
    dot.attr("node", shape="circle", style="filled", fillcolor="lightblue")

    # Compute depths using LONGEST path from roots (matches generation layers)
    # For DAGs, longest path = topological layer = correct visualization
    depths: Dict[int, int] = {}
    roots = [nid for nid in graph.nodes if len(parents(graph, nid)) == 0]

    # Initialize roots at depth 0
    for r in roots:
        depths[r] = 0

    # Process in topological order, taking max depth from parents + 1
    # Use modified BFS that updates depth to maximum seen
    from collections import deque

    queue = deque(roots)
    while queue:
        node_id = queue.popleft()
        current_depth = depths[node_id]
        for edge in children(graph, node_id):
            child_id = edge.dst
            new_depth = current_depth + 1
            if child_id not in depths or depths[child_id] < new_depth:
                depths[child_id] = new_depth
                queue.append(child_id)

    # Group by depth for subgraph ranking
    max_depth = max(depths.values()) if depths else 0
    levels: Dict[int, List[int]] = {d: [] for d in range(max_depth + 1)}
    for nid, d in depths.items():
        levels[d].append(nid)

    # Add nodes with same rank per level
    for depth in range(max_depth + 1):
        with dot.subgraph() as s:  # type: ignore
            s.attr(rank="same")
            for nid in sorted(levels[depth]):
                node = graph.nodes[nid]
                label = f"{nid}"
                if node.wait_after_ready > 0:
                    label += f"\\n({node.wait_after_ready:.1f}s)"
                s.node(str(nid), label)

    # Add edges with skip connection highlighting
    for src_id in graph.nodes:
        src_depth = depths.get(src_id, 0)
        for edge in children(graph, src_id):
            dst_depth = depths.get(edge.dst, 0)
            layer_span = dst_depth - src_depth

            if edge.is_history_parent:
                # History parent: red, bold
                if layer_span > 1:
                    # Skip connection + history parent: red, bold, dashed
                    dot.edge(
                        str(edge.src),
                        str(edge.dst),
                        color="red",
                        penwidth="2",
                        style="dashed",
                        label=f"H (+{layer_span})",
                    )
                else:
                    dot.edge(
                        str(edge.src),
                        str(edge.dst),
                        color="red",
                        penwidth="2",
                        label="H",
                    )
            else:
                if layer_span > 1:
                    # Skip connection: blue, dashed
                    dot.edge(
                        str(edge.src),
                        str(edge.dst),
                        color="blue",
                        style="dashed",
                        label=f"+{layer_span}",
                    )
                else:
                    # Regular edge: gray
                    dot.edge(str(edge.src), str(edge.dst), color="gray")

    # Render if path provided
    if output_path:
        dot.render(output_path, format=format, view=view, cleanup=True)

    return dot.source



"""Composite evaluator for running multiple evaluators in a single benchmark."""

from __future__ import annotations

from typing import Any, Dict, Optional, Set

from veeksha.evaluator.base import BaseEvaluator, EvaluationResult


class CompositeEvaluator(BaseEvaluator):
    """Fan-out wrapper that runs multiple evaluators over the same run."""

    def __init__(self, evaluators: list[BaseEvaluator]):
        if not evaluators:
            raise ValueError("CompositeEvaluator requires at least one evaluator.")
        super().__init__(
            config=evaluators[0].config, seed_manager=evaluators[0].seed_manager
        )
        self._evaluators = evaluators

    @property
    def _primary(self) -> BaseEvaluator:
        """Evaluator used for progress tracking and timeout filtering."""
        return self._evaluators[0]

    def register_request(
        self,
        request_id: int,
        session_id: int,
        dispatched_at: float,
        channels: Dict[Any, Any],
    ) -> None:
        for evaluator in self._evaluators:
            evaluator.register_request(
                request_id=request_id,
                session_id=session_id,
                dispatched_at=dispatched_at,
                channels=channels,
            )

    def record_request_completed(
        self,
        request_id: int,
        session_id: int,
        completed_at: float,
        response: Any,
        error: Optional[Exception] = None,
    ) -> None:
        for evaluator in self._evaluators:
            evaluator.record_request_completed(
                request_id=request_id,
                session_id=session_id,
                completed_at=completed_at,
                response=response,
                error=error,
            )

    def record_session_completed(
        self,
        session_id: int,
        completed_at: float,
        success: bool,
    ) -> None:
        for evaluator in self._evaluators:
            evaluator.record_session_completed(
                session_id=session_id, completed_at=completed_at, success=success
            )

    def finalize(self) -> EvaluationResult:
        results = [e.finalize() for e in self._evaluators]

        metrics: Dict[str, Any] = {}
        raw: Dict[str, Any] = {"evaluators": []}
        for idx, result in enumerate(results):
            key = result.evaluator_type or f"evaluator_{idx}"
            # Avoid collisions if evaluator_type repeats.
            if key in metrics:
                key = f"{key}_{idx}"
            metrics[key] = result.metrics
            raw["evaluators"].append(result.to_dict())

        return EvaluationResult(
            evaluator_type="composite",
            channel=None,
            metrics=metrics,
            raw_data=raw,
        )

    def save(self, output_dir: str) -> None:
        for evaluator in self._evaluators:
            evaluator.save(output_dir)

    # ---------------------------------------------------------------------
    # Progress / timeout hooks
    # ---------------------------------------------------------------------

    def get_streaming_metrics(self) -> Optional[Dict[str, Any]]:
        return self._primary.get_streaming_metrics()

    def get_completed_request_count(self) -> int:
        return self._primary.get_completed_request_count()

    def get_session_counts(self) -> tuple[int, int, int]:
        return self._primary.get_session_counts()

    def set_included_requests(self, request_ids: Set[int]) -> None:
        for evaluator in self._evaluators:
            evaluator.set_included_requests(request_ids)

    def get_registered_request_ids(self) -> Set[int]:
        primary = self._primary
        if hasattr(primary, "get_registered_request_ids"):
            return primary.get_registered_request_ids()  # type: ignore[no-any-return]
        return set()



from veeksha.config.generator.interval import (
    FixedIntervalGeneratorConfig,
)
from veeksha.generator.interval.base import BaseIntervalGenerator


class FixedIntervalGenerator(BaseIntervalGenerator):
    def __init__(self, config: FixedIntervalGeneratorConfig, rng: None):
        self.config = config

    def get_next_interval(self) -> float:
        return self.config.interval



from dataclasses import field

from veeksha.config.core.frozen_dataclass import frozen_dataclass


@frozen_dataclass(allow_from_file=True)
class RuntimeConfig:
    max_sessions: int = field(
        default=25,
        metadata={"help": "Maximum number of sessions to generate. -1 for unlimited."},
    )
    benchmark_timeout: int = field(
        default=300,
        metadata={"help": "Benchmark timeout in seconds."},
    )
    post_timeout_grace_seconds: int = field(
        default=-1,
        metadata={
            "help": "Grace period for in-flight requests after timeout. -1 waits for all, 0 exits immediately."
        },
    )
    num_dispatcher_threads: int = field(
        default=2,
        metadata={"help": "Number of threads for dispatching requests to workers."},
    )
    num_completion_threads: int = field(
        default=2,
        metadata={"help": "Number of threads for processing completed requests."},
    )
    num_client_threads: int = field(
        default=3,
        metadata={
            "help": "Number of async worker threads for making concurrent requests."
        },
    )



from dataclasses import field
from typing import Optional

from veeksha.config.client import (
    BaseClientConfig,
    OpenAIChatCompletionsClientConfig,
)
from veeksha.config.core.flat_dataclass import create_flat_dataclass
from veeksha.config.core.frozen_dataclass import frozen_dataclass
from veeksha.config.evaluator import (
    BaseEvaluatorConfig,
    PerformanceEvaluatorConfig,
)
from veeksha.config.generator.session import (
    BaseSessionGeneratorConfig,
    SyntheticSessionGeneratorConfig,
)
from veeksha.config.runtime import RuntimeConfig
from veeksha.config.server import BaseServerConfig
from veeksha.config.trace_recorder import TraceRecorderConfig
from veeksha.config.traffic import BaseTrafficConfig, RateTrafficConfig
from veeksha.config.wandb import WandbConfig
from veeksha.types import (
    ClientType,
    EvaluationType,
    SessionGeneratorType,
    TrafficType,
)


@frozen_dataclass(allow_from_file=True)
class BenchmarkConfig:
    output_dir: str = field(
        default="benchmark_output",
        metadata={
            "help": "Base directory for all benchmark outputs (traces, metrics, logs)"
        },
    )
    seed: int = field(
        default=42, metadata={"help": "Seed for the random number generator."}
    )
    session_generator: BaseSessionGeneratorConfig = field(
        default_factory=SyntheticSessionGeneratorConfig,
        metadata={
            "help": f"The session generator configuration for the benchmark. {SessionGeneratorType.help_str()}"
        },
    )
    traffic_scheduler: BaseTrafficConfig = field(
        default_factory=RateTrafficConfig,
        metadata={
            "help": f"The traffic scheduler configuration for the benchmark. {TrafficType.help_str()}"
        },
    )
    evaluators: list[BaseEvaluatorConfig] = field(
        default_factory=lambda: [PerformanceEvaluatorConfig()],
        metadata={"help": f"List of evaluators to run. {EvaluationType.help_str()}"},
    )
    client: BaseClientConfig = field(
        default_factory=OpenAIChatCompletionsClientConfig,
        metadata={
            "help": f"The client configuration for the benchmark. {ClientType.help_str()}"
        },
    )
    runtime: RuntimeConfig = field(
        default_factory=RuntimeConfig,
        metadata={"help": "The runtime configuration for the benchmark."},
    )
    trace_recorder: TraceRecorderConfig = field(
        default_factory=TraceRecorderConfig,
        metadata={
            "help": "Trace recorder configuration. Records dispatched requests (unlike the evaluator, which records them after completion)."
        },
    )
    server: Optional[BaseServerConfig] = field(
        default=None,
        metadata={
            "help": "Server configuration for managed servers. If set, client.model, client.api_key and client.api_base will be overwritten."
        },
    )
    wandb: WandbConfig = field(
        default_factory=WandbConfig,
        metadata={"help": "Weights & Biases logging configuration."},
    )

    @classmethod
    def create_from_cli_args(cls):
        """Create BenchmarkConfig instances from CLI

        Returns:
            List of BenchmarkConfig instances (single or
            multiple configs if YAML expands to multiple configurations)
        """
        flat_configs = create_flat_dataclass(cls).create_from_cli_args()
        instances = []
        for flat_config in flat_configs:
            instance = flat_config.reconstruct_original_dataclass()
            object.__setattr__(instance, "__flat_config__", flat_config)
            instances.append(instance)
        return instances

    def __post_init__(self):
        if not self.evaluators:
            raise ValueError("BenchmarkConfig.evaluators must be non-empty.")



"""E2E Test for External (Mock) Server Benchmark."""

import json
import threading
import time
from http.server import BaseHTTPRequestHandler, HTTPServer
from unittest.mock import MagicMock, patch

import pytest
import yaml

from veeksha.benchmark import manage_benchmark_run
from veeksha.config.benchmark import BenchmarkConfig
from veeksha.config.utils import create_class_from_dict
from veeksha.core.request_content import TextChannelRequestContent

SAMPLE_CONFIG_URL = "veeksha/sample_configs/managed_server.yml"

class MockOpenAIHandler(BaseHTTPRequestHandler):
    def do_POST(self):
        content_len = int(self.headers.get('Content-Length', 0))
        body = self.rfile.read(content_len)
        data = json.loads(body)
        
        # Determine number of tokens requested
        max_tokens = data.get("max_completion_tokens", 10)
        
        response = {
            "id": "chatcmpl-mock",
            "object": "chat.completion",
            "created": int(time.time()),
            "choices": [{
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": "mock " * max_tokens
                },
                "finish_reason": "stop"
            }],
            "usage": {
                "prompt_tokens": 10,
                "completion_tokens": max_tokens,
                "total_tokens": 10 + max_tokens
            }
        }
        
        self.send_response(200)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        self.wfile.write(json.dumps(response).encode('utf-8'))
    
    def log_message(self, format, *args):
        pass # Silence logs

@pytest.fixture
def mock_openai_server():
    server = HTTPServer(('localhost', 0), MockOpenAIHandler)
    thread = threading.Thread(target=server.serve_forever)
    thread.daemon = True
    thread.start()
    yield server
    server.shutdown()

# @pytest.mark.e2e
def test_external_server_benchmark(mock_openai_server, tmp_path) -> None:
    # 1. Load config
    with open(SAMPLE_CONFIG_URL, "r") as f:
        config_dict = yaml.safe_load(f)
    
    # 2. Modify config for test
    # Remove managed server config to force external server mode
    if "server" in config_dict:
        del config_dict["server"]
        
    config_dict["runtime"]["max_sessions"] = 5 
    config_dict["runtime"]["benchmark_timeout"] = 10
    config_dict["output_dir"] = str(tmp_path)
    
    # Point client to our mock server
    port = mock_openai_server.server_port
    if "client" not in config_dict:
        config_dict["client"] = {}
    config_dict["client"]["api_base"] = f"http://localhost:{port}/v1"
    config_dict["client"]["api_key"] = "dummy"
    config_dict["client"]["model"] = "mock-model"
    
    # 3. Create BenchmarkConfig
    benchmark_config = create_class_from_dict(BenchmarkConfig, config_dict)


    # 4. Mock tokenizer to avoid HF download
    with patch("veeksha.benchmark.build_hf_tokenizer_handle_from_model") as mock_build_tok_bench, \
         patch("veeksha.core.tokenizer.build_hf_tokenizer_handle_from_model") as mock_build_tok_core:
        
        mock_handle = MagicMock()
        mock_handle.encode.return_value = [1] * 10
        mock_handle.decode.return_value = "mock_text"
        mock_handle.count_tokens.return_value = 0
        mock_build_tok_bench.return_value = mock_handle
        mock_build_tok_core.return_value = mock_handle

        # Mock the channel generator to return a realistic TextChannelRequestContent
        mock_channel_gen = MagicMock()
        mock_channel_gen.generate_content.return_value = TextChannelRequestContent(
            input_text="mock prompt",
            target_output_tokens=1,
            target_prompt_tokens=5,
        )

        with patch(
            "veeksha.generator.channel.registry.ChannelGeneratorRegistry.get",
            return_value=mock_channel_gen,
        ):
            # 5. Run
            result = manage_benchmark_run(benchmark_config)
            
            # 6. Verify
            # Check result metrics: all sessions succeeded
            assert result.metrics["Successful Sessions"] == 5
            assert result.metrics["Number of Completed Requests"] > 0
            assert result.metrics["Error Rate"] == 0



"""E2E Test for Trace-based Benchmark (Mooncake Flavor)."""

import json
import os
import threading
import time
from http.server import BaseHTTPRequestHandler, HTTPServer
from unittest.mock import MagicMock, patch

import pytest
import yaml

from veeksha.benchmark import manage_benchmark_run
from veeksha.config.benchmark import BenchmarkConfig
from veeksha.config.utils import create_class_from_dict
from veeksha.generator.session.trace.prompt_builder import TracePromptBuilder
from veeksha.core.request_content import TextChannelRequestContent

SAMPLE_CONFIG_URL = "veeksha/sample_configs/managed_server.yml"

class MockOpenAIHandler(BaseHTTPRequestHandler):
    def do_POST(self):
        content_len = int(self.headers.get('Content-Length', 0))
        body = self.rfile.read(content_len)
        data = json.loads(body)
        
        # Determine number of tokens requested
        max_tokens = data.get("max_completion_tokens", 10)
        
        response = {
            "id": "chatcmpl-mock",
            "object": "chat.completion",
            "created": int(time.time()),
            "choices": [{
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": "mock " * max_tokens
                },
                "finish_reason": "stop"
            }],
            "usage": {
                "prompt_tokens": 10,
                "completion_tokens": max_tokens,
                "total_tokens": 10 + max_tokens
            }
        }
        
        self.send_response(200)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        self.wfile.write(json.dumps(response).encode('utf-8'))
    
    def log_message(self, format, *args):
        pass # Silence logs

@pytest.fixture
def mock_openai_server():
    server = HTTPServer(('localhost', 0), MockOpenAIHandler)
    thread = threading.Thread(target=server.serve_forever)
    thread.daemon = True
    thread.start()
    yield server
    server.shutdown()

# @pytest.mark.e2e
def test_trace_benchmark_mooncake(mock_openai_server, tmp_path) -> None:
    # 1. Create dummy corpus
    corpus_file = tmp_path / "corpus.txt"
    corpus_file.write_text("Hello world this is a test corpus line one.\nLine two for corpus content generation.\n")
    
    # 2. Create dummy trace
    trace_file = tmp_path / "trace.jsonl"
    trace_data = [
        {"session_id": 1, "input_length": 10, "output_length": 5, "new_input_length": 10, "hash_ids": [101, 102], "timestamp": 1000},
        {"session_id": 1, "input_length": 15, "output_length": 5, "new_input_length": 5, "hash_ids": [101, 102, 103], "timestamp": 1005},
        {"session_id": 2, "input_length": 10, "output_length": 5, "new_input_length": 10, "hash_ids": [101, 104], "timestamp": 1010},
    ]
    with open(trace_file, "w") as f:
        for item in trace_data:
            f.write(json.dumps(item) + "\n")
            
    # 3. Load sample config pattern
    with open(SAMPLE_CONFIG_URL, "r") as f:
        config_dict = yaml.safe_load(f)
    
    # 4. Modify config for trace test
    if "server" in config_dict:
        del config_dict["server"]
        
    config_dict["runtime"]["max_sessions"] = 2  # Matches number of unique sessions in trace
    config_dict["runtime"]["benchmark_timeout"] = 10
    config_dict["output_dir"] = str(tmp_path / "output")
    
    # Configure Session Generator as TRACE
    config_dict["session_generator"] = {
        "type": "trace",
        "trace_file": str(trace_file),
        "wrap_mode": False,
        "flavor": {
            "type": "mooncake_conv",
            "block_size": 4, # Small block size
            "corpus_file": str(corpus_file),
        }
    }
    
    # Point client to our mock server
    port = mock_openai_server.server_port
    if "client" not in config_dict:
        config_dict["client"] = {}
    config_dict["client"]["api_base"] = f"http://localhost:{port}/v1"
    config_dict["client"]["api_key"] = "dummy"
    config_dict["client"]["model"] = "mock-model"
    
    # 5. Create BenchmarkConfig
    benchmark_config = create_class_from_dict(BenchmarkConfig, config_dict)
    
    # 6. Mock dependencies
    with patch("veeksha.benchmark.build_hf_tokenizer_handle_from_model") as mock_build_tok_bench, \
         patch("veeksha.core.tokenizer.build_hf_tokenizer_handle_from_model") as mock_build_tok_core:

        mock_handle = MagicMock()
        mock_handle.encode.return_value = [1] * 10
        mock_handle.decode.return_value = "mock_text"
        mock_handle.count_tokens.return_value = 0
        mock_handle.get_vocab.return_value = ["a", "b"]
        mock_build_tok_bench.return_value = mock_handle
        mock_build_tok_core.return_value = mock_handle

        with patch.object(TracePromptBuilder, "build_from_hash_ids", return_value="mock prompt"):
            result = manage_benchmark_run(benchmark_config)

            # Verify basic metrics: sessions succeeded, no errors
            assert result.metrics.get("Successful Sessions", 0) == 2
            assert result.metrics.get("Number of Completed Requests", 0) >= 3
            assert result.metrics.get("Error Rate", 0) == 0



from veeksha.core.lazy_loader import _LazyLoader
from veeksha.types import SessionGraphType
from veeksha.types.base_registry import BaseRegistry


class SessionGraphGeneratorRegistry(BaseRegistry):
    @classmethod
    def get_key_from_str(cls, key_str: str) -> SessionGraphType:
        return SessionGraphType.from_str(key_str)  # type: ignore


SessionGraphGeneratorRegistry.register(
    SessionGraphType.LINEAR,
    _LazyLoader(
        "veeksha.generator.session_graph.linear",
        "LinearSessionGraphGenerator",
    ),
)

SessionGraphGeneratorRegistry.register(
    SessionGraphType.SINGLE_REQUEST,
    _LazyLoader(
        "veeksha.generator.session_graph.single_request",
        "SingleRequestSessionGraphGenerator",
    ),
)

SessionGraphGeneratorRegistry.register(
    SessionGraphType.BRANCHING,
    _LazyLoader(
        "veeksha.generator.session_graph.branching",
        "BranchingSessionGraphGenerator",
    ),
)



"""Unit tests for CLI commands."""

import subprocess

import pytest


@pytest.mark.unit
class TestCLI:
    """Test CLI commands work correctly."""

    def test_benchmark_help_command(self) -> None:
        """Test that benchmark help command works."""
        cmd = ["python", "-m", "veeksha.benchmark", "-h"]
        result = subprocess.run(cmd, capture_output=True, text=True)
        assert result.returncode == 0, "Help command failed"
        assert "usage:" in result.stdout.lower(), "Help output missing usage"
        # assert "--client-config-model" in result.stdout, "Help missing expected arguments"

    def test_capacity_search_help_command(self) -> None:
        """Test that capacity search help command works."""
        cmd = ["python", "-m", "veeksha.capacity_search", "-h"]
        result = subprocess.run(cmd, capture_output=True, text=True)
        assert result.returncode == 0, "Help command failed"
        assert "usage:" in result.stdout.lower(), "Help output missing usage"
        # assert "--slos" in result.stdout, "Help missing SLOs argument"
        assert "--max-iterations" in result.stdout, "Help missing max iterations argument"


from veeksha.slo.evaluator import SloCheckResult, SloEvaluationResult, SloEvaluator
from veeksha.slo.runner import evaluate_and_save_slos
from veeksha.slo.slo import BaseSlo, ConstantSlo, SloSet

__all__ = [
    "BaseSlo",
    "ConstantSlo",
    "SloSet",
    "SloEvaluator",
    "SloCheckResult",
    "SloEvaluationResult",
    "evaluate_and_save_slos",
]



from veeksha.core.lazy_loader import _LazyLoader
from veeksha.types import TrafficType
from veeksha.types.base_registry import BaseRegistry


class TrafficSchedulerRegistry(BaseRegistry):
    @classmethod
    def get_key_from_str(cls, key_str: str) -> TrafficType:
        return TrafficType.from_str(key_str)  # type: ignore


TrafficSchedulerRegistry.register(
    TrafficType.RATE,
    _LazyLoader(
        "veeksha.traffic.rate",
        "RateTrafficScheduler",
    ),
)

TrafficSchedulerRegistry.register(
    TrafficType.CONCURRENT,
    _LazyLoader(
        "veeksha.traffic.concurrent",
        "ConcurrentTrafficScheduler",
    ),
)



from dataclasses import field

from veeksha.config.core.frozen_dataclass import frozen_dataclass


@frozen_dataclass(allow_from_file=True)
class TraceRecorderConfig:
    """Configuration for request tracing"""

    enabled: bool = field(
        default=True, metadata={"help": "Enable recording of dispatched requests"}
    )
    include_content: bool = field(
        default=False,
        metadata={
            "help": "Include content of the request (channel blobs, history) in trace"
        },
    )



from dataclasses import field

from veeksha.config.core.base_poly_config import BasePolyConfig
from veeksha.config.core.frozen_dataclass import frozen_dataclass
from veeksha.config.generator.channel import (
    BaseChannelGeneratorConfig,
    TextChannelGeneratorConfig,
)
from veeksha.config.generator.session_graph import (
    BaseSessionGraphGeneratorConfig,
    LinearSessionGraphGeneratorConfig,
)
from veeksha.types import (
    ChannelModality,
    SessionGeneratorType,
    SessionGraphType,
    TraceFlavorType,
)


@frozen_dataclass(allow_from_file=True)
class BaseSessionGeneratorConfig(BasePolyConfig):
    pass


@frozen_dataclass
class SyntheticSessionGeneratorConfig(BaseSessionGeneratorConfig):
    session_graph: BaseSessionGraphGeneratorConfig = field(
        default_factory=LinearSessionGraphGeneratorConfig,
        metadata={
            "help": f"The generator for the session graphs. {SessionGraphType.help_str()}"
        },
    )
    channels: list[BaseChannelGeneratorConfig] = field(
        default_factory=lambda: [TextChannelGeneratorConfig()],
        metadata={
            "help": f"The modality channels for the content of each request. {ChannelModality.help_str()}"
        },
    )

    @classmethod
    def get_type(cls):
        return SessionGeneratorType.SYNTHETIC

    def __post_init__(self):
        channel_types = set([channel.get_type() for channel in self.channels])
        if len(channel_types) != len(self.channels):
            raise ValueError("All channel generators must have unique types")

        if not self.channels:
            raise ValueError("At least one channel generator must be specified")


@frozen_dataclass
class LmevalSessionGeneratorConfig(BaseSessionGeneratorConfig):
    tasks: list[str] = field(
        default_factory=lambda: ["hellaswag"],
        metadata={"help": "The lm-eval tasks to evaluate the model on."},
    )
    num_fewshot: int = field(
        default=1,
        metadata={"help": "The number of fewshot examples to use for the tasks."},
    )
    # NOTE: We intentionally do not expose a separate `limit` knob here.
    # Control total evaluated sessions via `runtime.max_sessions` (and wall time via
    # `runtime.benchmark_timeout`) to keep run termination consistent across workloads.

    @classmethod
    def get_type(cls):
        return SessionGeneratorType.LMEVAL

    def __post_init__(self):
        if not self.tasks:
            raise ValueError("LmevalSessionGeneratorConfig requires at least one task.")


# ----- Trace Flavor Configs -----


@frozen_dataclass
class BaseTraceFlavorConfig(BasePolyConfig):
    """Base config for trace flavors."""


@frozen_dataclass
class ClaudeCodeTraceFlavorConfig(BaseTraceFlavorConfig):
    """Context-cached trace flavor configuration."""

    # TODO global corpus file
    corpus_file: str = field(
        default="traces/corpus.txt",
        metadata={"help": "Path to corpus file for prompt padding"},
    )
    page_size: int = field(
        default=16,
        metadata={"help": "Number of unique tokens per session prefix"},
    )

    @classmethod
    def get_type(cls):
        return TraceFlavorType.CLAUDE_CODE


@frozen_dataclass
class MooncakeConvTraceFlavorConfig(BaseTraceFlavorConfig):
    """Mooncake conversation trace flavor configuration."""

    corpus_file: str = field(
        default="traces/corpus.txt",
        metadata={"help": "Path to corpus file for prompt padding"},
    )
    block_size: int = field(
        default=512,
        metadata={
            "help": "Number of tokens per hash id block. Only used for hash ids of first-in-session requests."
        },
    )

    @classmethod
    def get_type(cls):
        return TraceFlavorType.MOONCAKE_CONV


@frozen_dataclass
class RAGTraceFlavorConfig(BaseTraceFlavorConfig):
    """RAG trace flavor configuration."""

    num_documents: int = field(
        default=10,
        metadata={"help": "Number of top documents to include for warmup"},
    )

    @classmethod
    def get_type(cls):
        return TraceFlavorType.RAG


# ----- Trace Session Generator Config -----


@frozen_dataclass
class TraceSessionGeneratorConfig(BaseSessionGeneratorConfig):
    """Trace-driven session generator configuration."""

    trace_file: str = field(
        default="",
        metadata={"help": "Path to the JSONL trace file"},
    )
    wrap_mode: bool = field(
        default=True,
        metadata={"help": "Whether to wrap/loop over the trace indefinitely"},
    )
    flavor: BaseTraceFlavorConfig = field(
        default_factory=ClaudeCodeTraceFlavorConfig,
        metadata={"help": f"Trace flavor configuration. {TraceFlavorType.help_str()}"},
    )

    @classmethod
    def get_type(cls):
        return SessionGeneratorType.TRACE



import dataclasses
import typing
from typing import Any, Callable, TypeVar, overload

_T = TypeVar("_T")


@overload
def frozen_dataclass(_cls: type[_T]) -> type[_T]: ...


@overload
def frozen_dataclass(
    _cls: None = None, **kwargs: Any
) -> Callable[[type[_T]], type[_T]]: ...


@typing.dataclass_transform()
def frozen_dataclass(
    _cls: type[_T] | None = None, **kwargs: Any
) -> type[_T] | Callable[[type[_T]], type[_T]]:
    """
    A decorator that creates a frozen dataclass, allowing attribute modifications
    only during the __post_init__ method.

    Args:
        _cls: The class to decorate (for decorator syntax handling).
        **kwargs: Additional keyword arguments to pass to dataclasses.dataclass.
        allow_from_file (bool): If True, mark the dataclass as allowing configuration
            to be loaded from a file.
    """

    # Extract custom keyword arguments handled by this decorator
    allow_from_file = kwargs.pop("allow_from_file", False)

    def wrap(cls):
        # Apply dataclass with frozen=True
        datacls = dataclasses.dataclass(cls, frozen=True, **kwargs)  # type: ignore

        # Mark as allowing file-based configuration if requested
        if allow_from_file:
            setattr(datacls, "_allow_from_file", True)

        # Store the original __post_init__ if it exists
        original_post_init = getattr(cls, "__post_init__", None)

        # Define a new __post_init__ method
        def __post_init__(self):
            # Set the flag to allow attribute modifications
            object.__setattr__(self, "_in_post_init", True)
            try:
                if original_post_init:
                    original_post_init(self)
            finally:
                # Reset the flag after __post_init__ completes
                object.__setattr__(self, "_in_post_init", False)

        datacls.__post_init__ = __post_init__

        # Store the original __setattr__ method
        original_setattr = datacls.__setattr__

        # Define a new __setattr__ method
        def __setattr__(self, name, value):
            # Check if we are in __post_init__
            if getattr(self, "_in_post_init", False):
                object.__setattr__(self, name, value)
            else:
                original_setattr(self, name, value)

        datacls.__setattr__ = __setattr__

        return datacls

    # Support both @frozen_dataclass and @frozen_dataclass() syntax
    if _cls is None:
        return wrap
    return wrap(_cls)



"""Unit tests for SyntheticSessionGenerator."""

from typing import List

import pytest

from veeksha.config.generator.channel import TextChannelGeneratorConfig
from veeksha.config.generator.length import FixedLengthGeneratorConfig
from veeksha.config.generator.session import SyntheticSessionGeneratorConfig
from veeksha.config.generator.session_graph import LinearSessionGraphGeneratorConfig
from veeksha.core.seeding import SeedManager
from veeksha.core.tokenizer import TokenizerHandle, TokenizerProvider
from veeksha.generator.session.synthetic import SyntheticSessionGenerator
from veeksha.types import ChannelModality


def mock_tokenizer_handle() -> TokenizerHandle[str]:
    return TokenizerHandle(
        count_tokens=lambda x: len(x.split()),
        decode=lambda x: " ".join([str(t) for t in x]),
        encode=lambda x: [1] * len(x.split()), # Mock encode
        get_vocab=lambda: [1, 2, 3] # Mock vocab
    )

@pytest.fixture
def tokenizer_provider() -> TokenizerProvider:
    return TokenizerProvider(
        tokenizers={ChannelModality.TEXT: mock_tokenizer_handle()}
    )

@pytest.fixture
def linear_session_config() -> SyntheticSessionGeneratorConfig:
    return SyntheticSessionGeneratorConfig(
        session_graph=LinearSessionGraphGeneratorConfig(
            num_request_generator=FixedLengthGeneratorConfig(value=2)
        ),
        channels=[
            TextChannelGeneratorConfig(
                body_length_generator=FixedLengthGeneratorConfig(value=10),
                output_length_generator=FixedLengthGeneratorConfig(value=5)
            )
        ]
    )

@pytest.mark.unit
def test_generate_simple_session(tokenizer_provider, linear_session_config) -> None:
    seed_manager = SeedManager(seed=42)
    generator = SyntheticSessionGenerator(
        config=linear_session_config, 
        seed_manager=seed_manager,
        tokenizer_provider=tokenizer_provider
    )
    
    session = generator.generate_session()
    
    # Check session structure
    assert len(session.requests) == 2
    assert session.session_graph.nodes.keys() == {0, 1}
    
    # Check request content
    req0 = session.requests[0]
    assert ChannelModality.TEXT in req0.channels
    text_content = req0.channels[ChannelModality.TEXT]
    
    assert text_content.target_output_tokens == 5

@pytest.mark.unit
def test_generate_session_ids_increment(tokenizer_provider, linear_session_config) -> None:
    seed_manager = SeedManager(seed=42)
    generator = SyntheticSessionGenerator(
        config=linear_session_config, 
        seed_manager=seed_manager,
        tokenizer_provider=tokenizer_provider
    )
    
    s1 = generator.generate_session()
    s2 = generator.generate_session()
    
    assert s1.id == 0
    assert s2.id == 1
    
    # Check request IDs
    # s1 has 2 requests (req_ids 0, 1)
    # s2 has 2 requests (req_ids 2, 3)
    assert list(s1.requests.values())[0].id == 0
    assert list(s1.requests.values())[1].id == 1
    assert list(s2.requests.values())[0].id == 2



from veeksha.generator.channel.text import TextChannelGenerator
from veeksha.types import ChannelModality
from veeksha.types.base_registry import BaseRegistry


class ChannelGeneratorRegistry(BaseRegistry):
    @classmethod
    def get_key_from_str(cls, key_str: str) -> ChannelModality:
        return ChannelModality.from_str(key_str)  # type: ignore


ChannelGeneratorRegistry.register(ChannelModality.TEXT, TextChannelGenerator)



"""Unit tests for engine-specific server managers with advanced configs."""

import json
from textwrap import dedent

import pytest  # type: ignore[import]
import yaml

from veeksha.config.server import VllmServerConfig, VajraServerConfig
from veeksha.orchestration.vllm_server import VLLMServerManager
from veeksha.orchestration.vajra_server import VajraServerManager

pytestmark = pytest.mark.unit


def test_vllm_launch_command_with_advanced_configuration():
    config = VllmServerConfig(
        model="meta/test-model",
        host="0.0.0.0",
        port=9001,
        api_key="secret-key",
        tensor_parallel_size=2,
        dtype="bfloat16",
        max_model_len=8192,
        additional_args={
            "max-num-batched-tokens": 4096,
            "trust-remote-code": True,
            "disable-log-requests": False,
            "some_list": ["alpha", "beta"],
            "rope_scaling": {"type": "linear", "factor": 2.0},
        },
    )

    command = VLLMServerManager(config, output_dir="/tmp")._build_launch_command()

    # Base flags
    assert command[:3] == ["python", "-m", "vllm.entrypoints.openai.api_server"]
    assert command[command.index("--model") + 1] == "meta/test-model"
    assert command[command.index("--host") + 1] == "0.0.0.0"
    assert command[command.index("--port") + 1] == "9001"
    assert command[command.index("--api-key") + 1] == "secret-key"
    assert command[command.index("--tensor-parallel-size") + 1] == "2"
    assert command[command.index("--dtype") + 1] == "bfloat16"
    assert command[command.index("--max-model-len") + 1] == "8192"

    # Additional args
    assert command[command.index("--max-num-batched-tokens") + 1] == "4096"
    assert "--trust-remote-code" in command
    assert "--disable-log-requests" not in command

    list_start = command.index("--some_list")
    assert command[list_start : list_start + 3] == ["--some_list", "alpha", "beta"]

    rope_index = command.index("--rope-scaling")
    assert command[rope_index + 1] == json.dumps({"type": "linear", "factor": 2.0})


def test_vajra_launch_command_with_extra_cli_options():
    config = VajraServerConfig(
        model="meta/test-model",
        host="127.0.0.1",
        port=7777,
        api_key="another-secret",
        tensor_parallel_size=3,
        max_model_len=4096,
        additional_args={
            "pipeline_parallel_size": 4,
            "log_requests": True,
            "multi_value": ["foo", "bar"],
        },
    )

    command = VajraServerManager(config, output_dir="/tmp")._build_launch_command()

    assert command[:3] == ["python", "-m", "vajra_server.server"]
    assert command[command.index("--model") + 1] == "meta/test-model"
    assert command[command.index("--host") + 1] == "127.0.0.1"
    assert command[command.index("--port") + 1] == "7777"
    assert command[command.index("--api-key") + 1] == "another-secret"
    assert command[command.index("--tensor-parallel-size") + 1] == "3"
    assert command[command.index("--max-model-len") + 1] == "4096"

    assert command[command.index("--pipeline_parallel_size") + 1] == "4"
    assert "--log_requests" in command

    mv_first = command.index("--multi_value")
    assert command[mv_first : mv_first + 2] == ["--multi_value", "foo"]
    assert command[mv_first + 2 : mv_first + 4] == ["--multi_value", "bar"]


def test_server_config_additional_args_loaded_from_yaml(tmp_path):
    config_file = tmp_path / "server_config.yaml"
    config_file.write_text(
        dedent(
            """
            engine: vllm
            model: meta/demo-model
            port: 8100
            additional_args:
              trust-remote-code: true
              max-num-batched-tokens: 512
              rope_scaling:
                type: linear
                factor: 1.25
            """
        ).strip()
    )

    data = yaml.safe_load(config_file.read_text())
    # ServerConfig logic for creating from dict might be gone.
    # We should manually create VllmServerConfig if 'engine' is vllm.
    engine = data.pop("engine", "vllm")
    if engine == "vllm":
        server_config = VllmServerConfig(**data)
    else:
        server_config = VajraServerConfig(**data)
        
    command = VLLMServerManager(server_config, output_dir="/tmp")._build_launch_command()

    assert "--trust-remote-code" in command
    idx = command.index("--max-num-batched-tokens")
    assert command[idx + 1] == "512"

    rope_idx = command.index("--rope-scaling")
    assert json.loads(command[rope_idx + 1]) == {"type": "linear", "factor": 1.25}



import os
import threading
import time
from dataclasses import replace
from queue import Queue
from typing import Optional, Set

from veeksha.benchmark_utils import (
    _init_output_dir,
    _monitor_for_completion,
    build_evaluator,
    maybe_run_warmup,
)
from veeksha.client.registry import ClientRegistry
from veeksha.config.benchmark import BenchmarkConfig
from veeksha.core.seeding import SeedManager
from veeksha.core.thread_pool import ThreadPoolManager
from veeksha.core.tokenizer import (
    TokenizerProvider,
    build_hf_tokenizer_handle_from_model,
)
from veeksha.core.trace_recorder import TraceRecorder
from veeksha.generator.session.registry import SessionGeneratorRegistry
from veeksha.health import HealthChecker
from veeksha.logger import init_logger
from veeksha.orchestration import managed_server
from veeksha.traffic.registry import TrafficSchedulerRegistry
from veeksha.types import ChannelModality
from veeksha.wandb_integration import (
    maybe_finish_wandb_run,
    maybe_init_wandb_run,
    maybe_log_benchmark_artifacts,
    maybe_log_benchmark_scalars,
)
from veeksha.workers import CompletionWorker, DispatchWorker, PrefetchWorker
from veeksha.workers.client_runner import ClientRunnerManager
from veeksha.workers.prefetch import SharedSessionCounter

logger = init_logger(__name__)


def _run_main_loop(
    session_generator,
    traffic_scheduler,
    evaluator,
    client,
    runtime_config,
    trace_recorder=None,
    benchmark_start_time: Optional[float] = None,
) -> None:
    """Run the main benchmark loop with all workers."""
    logger.info("Starting main loop")
    if benchmark_start_time is None:
        benchmark_start_time = time.monotonic()

    client_queues = [Queue() for _ in range(runtime_config.num_client_threads)]
    output_queue = Queue()
    stop_event = threading.Event()
    generator_lock = threading.Lock()

    session_counter = SharedSessionCounter(max_sessions=runtime_config.max_sessions)

    client_runner = ClientRunnerManager(
        client=client,
        input_queues=client_queues,
        output_queue=output_queue,
        stop_event=stop_event,
    )

    pool_manager = ThreadPoolManager(stop_event=stop_event)

    pool_manager.create_pool(
        name="prefetch",
        worker_class=PrefetchWorker,
        worker_kwargs={
            "traffic_scheduler": traffic_scheduler,
            "session_generator": session_generator,
            "generator_lock": generator_lock,
            "session_counter": session_counter,
        },
        pool_size=1,
    )

    pool_manager.create_pool(
        name="dispatch",
        worker_class=DispatchWorker,
        worker_kwargs={
            "traffic_scheduler": traffic_scheduler,
            "client_queues": client_queues,
            "evaluator": evaluator,
            "trace_recorder": trace_recorder,
        },
        pool_size=runtime_config.num_dispatcher_threads,
    )

    pool_manager.create_pool(
        name="completion",
        worker_class=CompletionWorker,
        worker_kwargs={
            "output_queue": output_queue,
            "traffic_scheduler": traffic_scheduler,
            "evaluator": evaluator,
        },
        pool_size=runtime_config.num_completion_threads,
    )

    if trace_recorder:
        trace_recorder.start()

    client_runner.start()
    pool_manager.start_all()

    logger.info(
        f"Started {pool_manager.get_total_thread_count()} worker threads "
        f"and {client_runner.get_worker_count()} client workers"
    )

    benchmark_start = benchmark_start_time
    benchmark_timeout = runtime_config.benchmark_timeout
    timeout_triggered = False
    pre_timeout_request_ids: Set[str] = set()

    try:
        pending_in_flight = _monitor_for_completion(
            traffic_scheduler,
            evaluator,
            pool_manager,
            benchmark_start,
            benchmark_timeout,
            timeout_triggered,
            pre_timeout_request_ids,
            max_sessions=runtime_config.max_sessions,
            post_timeout_grace_seconds=runtime_config.post_timeout_grace_seconds,
        )
    except KeyboardInterrupt:
        logger.info("Interrupted, stopping")
        pending_in_flight = set()

    stop_event.set()
    pool_manager.join_pool("prefetch", timeout=1.0)
    pool_manager.join_pool("dispatch", timeout=1.0)

    if trace_recorder:
        trace_recorder.stop()

    logger.info("Stopping client runner...")
    client_runner.stop()
    if not pending_in_flight:
        client_runner.wait()

    for _ in range(runtime_config.num_completion_threads):
        output_queue.put(None)
    pool_manager.join_pool("completion", timeout=1.0)


def _run_benchmark(
    benchmark_config: BenchmarkConfig,
):
    """Run the benchmark and return evaluation results.

    Args:
        benchmark_config: The benchmark configuration.

    Returns:
        EvaluationResult from the evaluator.
    """

    seed_manager = SeedManager(benchmark_config.seed)

    # get session generator
    model_name = benchmark_config.client.model
    tokenizer_provider = TokenizerProvider(
        {ChannelModality.TEXT: build_hf_tokenizer_handle_from_model(model_name)},
        model_name=model_name,
    )
    append_min_tokens_instruction = False
    if (
        hasattr(benchmark_config.client, "use_min_tokens_prompt_fallback")
        and benchmark_config.client.use_min_tokens_prompt_fallback  # type: ignore
    ):
        append_min_tokens_instruction = True
        logger.info(
            "Min tokens prompt fallback enabled in config. "
            "Will append instructions to prompts for minimum token control."
        )

    session_generator_kwargs = {
        "config": benchmark_config.session_generator,
        "seed_manager": seed_manager,
        "tokenizer_provider": tokenizer_provider,
    }

    # lm-eval uses runtime.max_sessions as the only sample-size knob.
    if (
        benchmark_config.session_generator.get_type()
        == SessionGeneratorRegistry.get_key_from_str("lmeval")
    ):
        session_generator_kwargs["max_sessions"] = benchmark_config.runtime.max_sessions

    if (
        benchmark_config.session_generator.get_type()
        == SessionGeneratorRegistry.get_key_from_str("synthetic")
    ):
        session_generator_kwargs["append_min_tokens_instruction"] = (
            append_min_tokens_instruction
        )

    session_generator = SessionGeneratorRegistry.get(
        benchmark_config.session_generator.get_type(),
        **session_generator_kwargs,
    )

    # get traffic scheduler, client
    traffic_scheduler = TrafficSchedulerRegistry.get(
        benchmark_config.traffic_scheduler.get_type(),
        config=benchmark_config.traffic_scheduler,
        seed_manager=seed_manager,
    )

    client = ClientRegistry.get(
        benchmark_config.client.get_type(),
        config=benchmark_config.client,
        tokenizer_provider=tokenizer_provider,
    )

    # some session generators might define a warmup phase
    maybe_run_warmup(session_generator, client)

    benchmark_start_time = time.monotonic()
    traffic_scheduler.reset_reference_time()

    # get evaluator
    evaluator = build_evaluator(
        benchmark_config,
        seed_manager=seed_manager,
        session_generator=session_generator,
        benchmark_start_time=benchmark_start_time,
    )

    # trace recorder
    trace_recorder = None
    if benchmark_config.trace_recorder.enabled:
        # ensure output dirs exists for traces
        os.makedirs(f"{benchmark_config.output_dir}/traces", exist_ok=True)
        trace_recorder = TraceRecorder(
            f"{benchmark_config.output_dir}/traces",
            benchmark_start_time,
            benchmark_config.trace_recorder,
        )
        trace_recorder.start()

    os.makedirs(f"{benchmark_config.output_dir}/metrics", exist_ok=True)

    try:
        _run_main_loop(
            session_generator=session_generator,
            traffic_scheduler=traffic_scheduler,
            evaluator=evaluator,
            client=client,
            runtime_config=benchmark_config.runtime,
            trace_recorder=trace_recorder,
            benchmark_start_time=benchmark_start_time,
        )
    finally:
        if trace_recorder:
            trace_recorder.stop()

    logger.info("Finalizing evaluator...")
    # finalize and save results
    result = evaluator.finalize()

    evaluator.save(f"{benchmark_config.output_dir}/metrics")

    # health checks
    logger.info("Running health checks...")
    health_checker = HealthChecker(
        trace_file=f"{benchmark_config.output_dir}/traces/dispatch_trace.jsonl",
        metrics_file=f"{benchmark_config.output_dir}/metrics/request_level_metrics.jsonl",
        benchmark_config=benchmark_config,
    )
    health_checker.run_and_save(
        f"{benchmark_config.output_dir}/health_check_results.txt"
    )

    return result


def manage_benchmark_run(
    benchmark_config: BenchmarkConfig,
):
    """Run a benchmark, handling optional server orchestration.

    1. If server config exists: spin up server, update client config, run benchmark
    2. If no server config: run benchmark directly

    Args:
        benchmark_config: The benchmark configuration.

    Returns:
        EvaluationResult from the evaluator.
    """
    logger.info("Running benchmark with config:\n%s", benchmark_config)

    _init_output_dir(benchmark_config)

    if benchmark_config.server is not None:
        logger.info(f"Launching {benchmark_config.server.engine} server...")

        with managed_server(
            benchmark_config.server, output_dir=benchmark_config.output_dir
        ) as server_info:
            logger.info(f"Server ready at {server_info['api_base']}")

            # server dictates client
            updated_client_config = replace(
                benchmark_config.client,
                api_base=server_info["api_base"],
                api_key=server_info["api_key"],
                model=benchmark_config.server.model,
            )
            updated_benchmark_config = replace(
                benchmark_config,
                client=updated_client_config,
                server=None,
            )

            maybe_init_wandb_run(updated_benchmark_config, run_kind="benchmark")
            try:
                result = _run_benchmark(updated_benchmark_config)
                maybe_log_benchmark_scalars(updated_benchmark_config.output_dir)
                maybe_log_benchmark_artifacts(updated_benchmark_config)
                return result
            finally:
                maybe_finish_wandb_run(updated_benchmark_config.output_dir)
                logger.info("Server shutting down...")
    else:
        maybe_init_wandb_run(benchmark_config, run_kind="benchmark")
        try:
            result = _run_benchmark(benchmark_config)
            maybe_log_benchmark_scalars(benchmark_config.output_dir)
            maybe_log_benchmark_artifacts(benchmark_config)
            return result
        finally:
            maybe_finish_wandb_run(benchmark_config.output_dir)


if __name__ == "__main__":
    from veeksha.cli.benchmarks import main

    main()



from dataclasses import field
from typing import Optional

from veeksha.config.core.base_poly_config import BasePolyConfig
from veeksha.config.core.frozen_dataclass import frozen_dataclass
from veeksha.types import SloType

SUPPORTED_SLO_METRICS = {"ttfc", "tbc", "tpot", "e2e"}


@frozen_dataclass
class BaseSloConfig(BasePolyConfig):
    """Base class for a single SLO definition."""

    percentile: float = field(
        default=0.99,
        metadata={"help": "Percentile at which to evaluate the SLO (0.0-1.0)"},
    )
    name: Optional[str] = field(
        default=None, metadata={"help": "Human-readable name for this SLO"}
    )

    def __post_init__(self) -> None:
        if not 0.0 <= self.percentile <= 1.0:
            raise ValueError(
                f"percentile must be between 0.0 and 1.0, got {self.percentile}"
            )


@frozen_dataclass
class ConstantSloConfig(BaseSloConfig):
    """SLO with a fixed constant value threshold."""

    metric: str = field(
        default="ttfc",
        metadata={
            "help": f"The metric key this SLO applies to. Available: {', '.join(sorted(SUPPORTED_SLO_METRICS))}."
        },
    )

    value: float = field(
        default=-1.0,
        metadata={
            "help": "The constant value for the SLO. If a percentage, from 0 to 1. If a time, in seconds."
        },
    )

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.metric not in SUPPORTED_SLO_METRICS:
            raise ValueError(
                f"ConstantSLO: metric '{self.metric}' is not supported. "
                f"Supported: {sorted(SUPPORTED_SLO_METRICS)}"
            )
        if self.value <= 0.0:
            raise ValueError("ConstantSLO: value must be specified and must be > 0")

    @classmethod
    def get_type(cls) -> SloType:
        return SloType.CONSTANT



"""Workers for the new Veeksha framework."""

from veeksha.core.context import WorkerContext
from veeksha.workers.completion import CompletionWorker
from veeksha.workers.dispatch import DispatchWorker
from veeksha.workers.prefetch import PrefetchWorker

__all__ = [
    "PrefetchWorker",
    "DispatchWorker",
    "CompletionWorker",
    "WorkerContext",
]



from __future__ import annotations

import json
import os
from typing import Any, Dict, Optional, Sequence, Set

from veeksha.config.slo import BaseSloConfig
from veeksha.logger import init_logger
from veeksha.slo.evaluator import SloEvaluationResult, SloEvaluator
from veeksha.slo.slo import SloSet

logger = init_logger(__name__)


def load_request_level_metrics_jsonl(
    path: str, *, keys: Optional[Set[str]] = None
) -> Dict[str, Any]:
    """Load request-level metrics JSONL into a dict-of-lists.

    Args:
        path: Path to request_level_metrics.jsonl
        keys: If provided, only these keys are loaded (reduces memory/CPU).

    Returns:
        Dict mapping metric keys to list-of-values.
    """
    metrics: Dict[str, Any] = {}
    if not os.path.exists(path):
        return metrics

    with open(path, "r", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if not line:
                continue
            row = json.loads(line)
            if not isinstance(row, dict):
                continue
            if keys is None:
                for key, value in row.items():
                    metrics.setdefault(key, []).append(value)
            else:
                for key in keys:
                    if key in row:
                        metrics.setdefault(key, []).append(row[key])
    return metrics


def evaluate_and_save_slos(
    slo_configs: Sequence[BaseSloConfig],
    metrics_dir: str,
) -> Optional[SloEvaluationResult]:
    """Evaluate configured SLOs and persist results to the metrics directory.

    Persists:
    - `slo_results.json`: machine-readable summary for meta-benchmarks

    Args:
        slo_configs: SLO configs to evaluate.
        metrics_dir: Directory containing request-level metrics and where SLO
            results should be written.

    Returns:
        SloEvaluationResult if SLOs were evaluated, else None.
    """
    if not slo_configs:
        return None

    os.makedirs(metrics_dir, exist_ok=True)

    request_metrics_path = os.path.join(metrics_dir, "request_level_metrics.jsonl")
    request_level_metrics = load_request_level_metrics_jsonl(
        request_metrics_path, keys=_get_slo_metric_keys(slo_configs)
    )
    if not request_level_metrics:
        logger.warning(
            "SLO evaluation requested but request-level metrics file is missing/empty: %s",
            request_metrics_path,
        )

    slo_set = SloSet(list(slo_configs))
    evaluator = SloEvaluator(slo_set)
    evaluation_result = evaluator.evaluate(request_level_metrics)

    out_path = os.path.join(metrics_dir, "slo_results.json")
    with open(out_path, "w", encoding="utf-8") as f:
        json.dump(evaluation_result.to_dict(), f, indent=2)

    return evaluation_result


def _get_slo_metric_keys(slo_configs: Sequence[BaseSloConfig]) -> Set[str]:
    keys: Set[str] = set()
    for slo in slo_configs:
        metric = getattr(slo, "metric", None)
        if isinstance(metric, str):
            if metric == "e2e":
                keys.add("end_to_end_latency")
            else:
                keys.add(metric)
    return keys



import json
import os

import pytest

from veeksha.sweep_summary import write_sweep_summary


@pytest.mark.unit
def test_write_sweep_summary_aggregates_decode_window_metrics(tmp_path) -> None:
    base = tmp_path / "decode_sweep"
    base.mkdir()

    run1 = base / "run1"
    run2 = base / "run2"
    (run1 / "metrics").mkdir(parents=True)
    (run2 / "metrics").mkdir(parents=True)

    # Minimal config.yml for context length inference.
    (run1 / "config.yml").write_text(
        """
traffic_scheduler:
  type: concurrent
  target_concurrent_sessions: 4
session_generator:
  channels:
    - type: text
      body_length_generator:
        type: fixed
        value: 512
""".lstrip()
    )
    (run2 / "config.yml").write_text(
        """
traffic_scheduler:
  type: concurrent
  target_concurrent_sessions: 8
session_generator:
  channels:
    - type: text
      body_length_generator:
        type: fixed
        value: 1024
""".lstrip()
    )

    # Prefill stats with single group (helps inference).
    (run1 / "metrics" / "prefill_stats.json").write_text(
        json.dumps({"groups": {"512": {"count": 1}}})
    )
    (run2 / "metrics" / "prefill_stats.json").write_text(
        json.dumps({"groups": {"1024": {"count": 1}}})
    )

    # Decode window metrics (what we aggregate).
    (run1 / "metrics" / "decode_window_metrics.json").write_text(
        json.dumps(
            {
                "config": {"min_active_requests": 4},
                "window": {"start": 0.1, "end": 0.2},
                "tbc_in_window_stats": {"count": 10, "mean": 0.01},
            }
        )
    )
    (run2 / "metrics" / "decode_window_metrics.json").write_text(
        json.dumps(
            {
                "config": {"min_active_requests": 8},
                "window": {"start": 0.1, "end": 0.2},
                "tbc_in_window_stats": {"count": 20, "mean": 0.02},
            }
        )
    )

    # Optional standard artifacts (ensure sweep_summary includes them when present).
    (run1 / "metrics" / "summary_stats.json").write_text(
        json.dumps({"Number of Requests": 10, "Number of Completed Requests": 10, "Error Rate": 0.0})
    )
    (run2 / "metrics" / "summary_stats.json").write_text(
        json.dumps({"Number of Requests": 20, "Number of Completed Requests": 20, "Error Rate": 0.0})
    )
    (run1 / "metrics" / "throughput_metrics.json").write_text(
        json.dumps({"tpot_based_throughput": 100.0, "tbc_based_throughput": 90.0})
    )
    (run2 / "metrics" / "throughput_metrics.json").write_text(
        json.dumps({"tpot_based_throughput": 200.0, "tbc_based_throughput": 180.0})
    )
    (run1 / "metrics" / "slo_results.json").write_text(
        json.dumps({"all_slos_met": True, "results": []})
    )
    (run2 / "metrics" / "slo_results.json").write_text(
        json.dumps({"all_slos_met": True, "results": []})
    )

    written = write_sweep_summary(
        str(base),
        [str(run1), str(run2)],
    )

    assert os.path.exists(written["sweep_manifest"])
    assert os.path.exists(written["sweep_summary"])
    assert os.path.exists(written["sweep_summary_csv"])
    assert os.path.exists(written["decode_stats"])
    assert os.path.exists(written["decode_stats_csv"])

    decode = json.loads((base / "decode_stats.json").read_text())
    assert "512_4" in decode
    assert "1024_8" in decode
    assert decode["512_4"]["tbc_in_window_stats"]["mean"] == pytest.approx(0.01)
    assert decode["1024_8"]["tbc_in_window_stats"]["mean"] == pytest.approx(0.02)





from abc import abstractmethod
from typing import Optional

from veeksha.config.generator.session import BaseSessionGeneratorConfig
from veeksha.core.seeding import SeedManager
from veeksha.core.session import Session


class BaseSessionGenerator:
    def __init__(
        self,
        config: BaseSessionGeneratorConfig,
        seed_manager: Optional[SeedManager] = None,
    ):
        self.config = config
        self.seed_manager = seed_manager or SeedManager(0)

    @abstractmethod
    def generate_session(self) -> Session:
        pass

    @abstractmethod
    def capacity(self) -> int:
        """Total number of sessions producible if finite; -1 if unbounded."""



import asyncio

import pytest

from veeksha.client import ClientRegistry
from veeksha.config.client import OpenAIRouterClientConfig
from veeksha.core.request import Request
from veeksha.core.request_content import TextChannelRequestContent
from veeksha.core.response import RequestResult
from veeksha.core.tokenizer import TokenizerHandle, TokenizerProvider
from veeksha.types import ChannelModality


@pytest.mark.unit
def test_openai_router_client_routes_per_request_api_mode() -> None:
    tokenizer_handle = TokenizerHandle(
        count_tokens=lambda text: len(str(text).split()),
        decode=lambda token_ids: "",
        encode=lambda text: [0] * len(str(text).split()),
    )
    tokenizer_provider = TokenizerProvider({ChannelModality.TEXT: tokenizer_handle})

    config = OpenAIRouterClientConfig(
        api_base="http://example.com/v1",
        api_key="",
        model="dummy",
    )
    client = ClientRegistry.get(
        config.get_type(),
        config=config,
        tokenizer_provider=tokenizer_provider,
    )

    called = {"chat": 0, "completions": 0}

    async def _fake_chat_send_request(
        *,
        request: Request,
        session_id: int,
        session_total_requests: int = 1,
    ) -> RequestResult:
        called["chat"] += 1
        return RequestResult(
            request_id=request.id,
            session_id=session_id,
            session_total_requests=session_total_requests,
            success=True,
            client_completed_at=0.0,
        )

    async def _fake_completions_send_request(
        *,
        request: Request,
        session_id: int,
        session_total_requests: int = 1,
    ) -> RequestResult:
        called["completions"] += 1
        return RequestResult(
            request_id=request.id,
            session_id=session_id,
            session_total_requests=session_total_requests,
            success=True,
            client_completed_at=0.0,
        )

    client._chat_client.send_request = _fake_chat_send_request  # type: ignore[attr-defined]
    client._completions_client.send_request = (  # type: ignore[attr-defined]
        _fake_completions_send_request
    )

    chat_req = Request(
        id=1,
        channels={
            ChannelModality.TEXT: TextChannelRequestContent(
                input_text="hello",
                target_output_tokens=1,
            )
        },
        metadata={"api_mode": "chat"},
    )
    completions_req = Request(
        id=2,
        channels={
            ChannelModality.TEXT: TextChannelRequestContent(
                input_text="hello",
                target_output_tokens=1,
            )
        },
        metadata={"api_mode": "completions"},
    )

    asyncio.run(client.send_request(chat_req, session_id=0, session_total_requests=1))
    asyncio.run(
        client.send_request(completions_req, session_id=0, session_total_requests=1)
    )

    assert called["chat"] == 1
    assert called["completions"] == 1


@pytest.mark.unit
def test_openai_router_client_uses_separate_max_tokens_params() -> None:
    """Verify the router uses different max_tokens_param for each endpoint.

    The /chat/completions endpoint uses max_tokens_param (default: max_completion_tokens),
    while /completions uses completions_max_tokens_param (default: max_tokens).
    """
    tokenizer_handle = TokenizerHandle(
        count_tokens=lambda text: len(str(text).split()),
        decode=lambda token_ids: "",
        encode=lambda text: [0] * len(str(text).split()),
    )
    tokenizer_provider = TokenizerProvider({ChannelModality.TEXT: tokenizer_handle})

    config = OpenAIRouterClientConfig(
        api_base="http://example.com/v1",
        api_key="",
        model="dummy",
        max_tokens_param="max_completion_tokens",  # for chat
        completions_max_tokens_param="max_tokens",  # for completions
    )
    client = ClientRegistry.get(
        config.get_type(),
        config=config,
        tokenizer_provider=tokenizer_provider,
    )

    # Verify the chat client uses the chat-specific max_tokens_param
    assert client._chat_client.config.max_tokens_param == "max_completion_tokens"

    # Verify the completions client uses the completions-specific max_tokens_param
    assert client._completions_client.config.max_tokens_param == "max_tokens"



"""RAG trace flavor generator with warmup support."""

from typing import Any, List, Optional

import pandas as pd

from veeksha.config.generator.session import (
    RAGTraceFlavorConfig,
    TraceSessionGeneratorConfig,
)
from veeksha.core.seeding import SeedManager
from veeksha.core.session import Session
from veeksha.core.tokenizer import TokenizerProvider
from veeksha.generator.session.trace.base_flavor import (
    TraceFlavorGeneratorBase,
)


class RAGTraceFlavorGenerator(TraceFlavorGeneratorBase):
    """RAG trace flavor generator with warmup sessions per document."""

    def __init__(
        self,
        config: TraceSessionGeneratorConfig,
        flavor_config: RAGTraceFlavorConfig,
        seed_manager: SeedManager,
        tokenizer_provider: TokenizerProvider,
    ):
        super().__init__(config, flavor_config, seed_manager, tokenizer_provider)
        self.flavor_config = flavor_config
        self._warmup_sessions: Optional[List[Session]] = None

        doc_counts = self.trace_df["doc_id"].value_counts()

        requested_docs = max(int(flavor_config.num_documents), 1)
        self._top_doc_ids: List[Any] = doc_counts.nlargest(
            requested_docs
        ).index.tolist()
        self._warmup_count = len(self._top_doc_ids)

        filtered_df = (
            self.trace_df[self.trace_df["doc_id"].isin(self._top_doc_ids)]
            .reset_index(drop=True)
            .copy()
        )
        if filtered_df.empty:
            raise ValueError("No trace rows remain after filtering to top documents. ")

        session_start = self._warmup_count
        filtered_df["session_id"] = range(
            session_start, session_start + len(filtered_df)
        )

        self._warmup_source_df = self._build_warmup_source(filtered_df)
        # shuffle operational trace to emulate historical behavior
        self.trace_df = self._shuffle_sessions(filtered_df)
        # ensure generated sessions follow warmup id range
        self._session_offset = session_start
        self._current_session_id = self._session_offset

    @property
    def required_columns(self) -> List[str]:
        return [
            "doc_id",
            "prompt_text",
            "input_length",
            "output_length",
        ]

    def _validate_trace(self):
        """Ensure the trace contains required columns and session ids."""
        super()._validate_trace()

        if "session_id" not in self.trace_df.columns:
            self.trace_df = self.trace_df.reset_index(drop=True)
            self.trace_df["session_id"] = range(len(self.trace_df))

    def get_warmup_sessions(self) -> List[Session]:
        """Generate warmup sessions-one per selected document."""
        if self._warmup_sessions is not None:
            return self._warmup_sessions

        sessions: List[Session] = []
        for session_id, (_, row) in enumerate(self._warmup_source_df.iterrows()):
            sessions.append(self._build_single_request_session(session_id, row))

        self._warmup_sessions = sessions
        # reshuffle main trace before benchmark sessions start
        self.trace_df = self._shuffle_sessions(self.trace_df)
        self._session_groups = None
        return self._warmup_sessions

    def prepare_session(self, group: pd.DataFrame) -> Session:
        """Prepare a single-request RAG session from the trace row."""
        row = group.iloc[0]
        session_id = self._next_session_id()
        return self._build_single_request_session(session_id, row)

    def wrap(self) -> pd.DataFrame:
        """Wrap trace for new epoch with refreshed session order."""
        df = self.trace_df.copy()
        max_sid = int(df["session_id"].max()) if not df.empty else self._session_offset
        df["session_id"] = df["session_id"] + max_sid + 1
        return self._shuffle_sessions(df)

    def _build_warmup_source(self, filtered_df: pd.DataFrame) -> pd.DataFrame:
        """Capture a deterministic row per doc_id for warmup sessions."""
        rows = []
        for doc_id in self._top_doc_ids:
            doc_rows = filtered_df[filtered_df["doc_id"] == doc_id]
            if not doc_rows.empty:
                rows.append(doc_rows.iloc[0])
        if rows:
            warmup_df = pd.DataFrame(rows).reset_index(drop=True)
        else:
            warmup_df = pd.DataFrame(columns=filtered_df.columns)

        warmup_df["session_id"] = range(len(warmup_df))
        return warmup_df

    def _build_single_request_session(self, session_id: int, row: pd.Series) -> Session:
        """Convert a single trace row into a Veeksha Session."""
        wait_time = 0.0
        request = self._create_text_request(
            node_id=0,
            prompt_text=str(row["prompt_text"]),
            target_output_tokens=int(row["output_length"]),
            wait_after_ready=wait_time,
            parent_node=None,
            target_prompt_tokens=int(row["input_length"]),
        )
        session_graph = self._build_linear_session_graph(1, [wait_time])
        return Session(
            id=session_id,
            session_graph=session_graph,
            requests={0: request},
        )



from veeksha.config.generator.session_graph import (
    SingleRequestSessionGraphGeneratorConfig,
)
from veeksha.core.seeding import SeedManager
from veeksha.core.session_graph import SessionGraph, SessionNode, add_node
from veeksha.generator.session_graph.base import BaseSessionGraphGenerator


class SingleRequestSessionGraphGenerator(BaseSessionGraphGenerator):
    def __init__(
        self,
        config: SingleRequestSessionGraphGeneratorConfig,
        seed_manager: SeedManager,
    ):
        self.config = config
        self.seed_manager = seed_manager

    def generate_session_graph(self) -> SessionGraph:
        session_graph = SessionGraph()
        add_node(session_graph, SessionNode(id=0, wait_after_ready=0))
        return session_graph



"""Client for OpenAI-compatible Completions API."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional

import httpx

from veeksha.client.openai_base import OpenAIBaseClient
from veeksha.core.request import Request
from veeksha.core.response import ChannelResponse, RequestResult
from veeksha.core.tokenizer import TokenizerProvider
from veeksha.logger import init_logger
from veeksha.types import ChannelModality

if TYPE_CHECKING:
    from veeksha.config.client import BaseClientConfig

logger = init_logger(__name__)


class OpenAICompletionsClient(OpenAIBaseClient):
    """Async client for OpenAI `/completions` API using httpx.

    This client always uses the completions endpoint (non-streaming). It is
    intended for servers that only expose completions, or workloads that require
    provider logprobs (e.g., lm-eval loglikelihood requests).
    """

    def __init__(
        self,
        config: BaseClientConfig,
        tokenizer_provider: TokenizerProvider,
    ) -> None:
        super().__init__(config=config, tokenizer_provider=tokenizer_provider)

        append = str(self.config.address_append_value)
        if append.endswith("chat/completions"):
            completions_path = append.replace("chat/completions", "completions")
        else:
            completions_path = "completions"
        self.completions_address = str(self.api_base) + completions_path

    async def send_request(
        self,
        request: Request,
        session_id: int,
        session_total_requests: int = 1,
    ) -> RequestResult:
        """Send a request to the OpenAI `/completions` endpoint (non-streaming)."""
        return await self._send_completions_request(
            request=request,
            session_id=session_id,
            session_total_requests=session_total_requests,
        )

    async def _send_completions_request(
        self,
        request: Request,
        session_id: int,
        session_total_requests: int,
    ) -> RequestResult:
        """Execute the HTTP request against `/completions` and parse the response."""
        timeout = self.config.request_timeout

        prompt_text = ""
        max_tokens_limit = None
        if ChannelModality.TEXT in request.channels:
            text_content = request.channels[ChannelModality.TEXT]
            prompt_text = text_content.input_text  # type: ignore
            max_tokens_limit = text_content.target_output_tokens  # type: ignore

        sampling_params = self._get_sampling_params(request)
        body: Dict[str, Any] = {
            "model": self.config.model,
            "prompt": prompt_text,
            "stream": False,
        }
        body.update(sampling_params)
        body["stream"] = False

        # Max tokens handling:
        # - Prefer explicit per-request sampling params
        # - Otherwise use the configured max_tokens_param
        max_tokens_param = getattr(self.config, "max_tokens_param", None)
        has_explicit_max_tokens = "max_tokens" in body
        has_configured_max_tokens = bool(max_tokens_param and max_tokens_param in body)
        if (
            max_tokens_limit is not None
            and int(max_tokens_limit) > 0
            and max_tokens_param
            and not has_explicit_max_tokens
            and not has_configured_max_tokens
        ):
            body[max_tokens_param] = max_tokens_limit

        headers = {
            "Authorization": f"Bearer {self.config.api_key}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

        error_msg: Optional[str] = None
        error_code: Optional[int] = None
        generated_text = ""
        completion_text = ""
        logprobs: Any = None

        start_time = time.monotonic()
        try:
            client = self._get_client()
            response = await client.post(
                self.completions_address, json=body, headers=headers, timeout=timeout
            )
            response.raise_for_status()
            data = response.json()
            choices = data.get("choices") or []
            if choices:
                first = choices[0] if isinstance(choices[0], dict) else {}
                generated_text = first.get("text", "") or ""
                logprobs = first.get("logprobs")
        except httpx.HTTPStatusError as e:
            error_code = e.response.status_code if e.response else 500
            error_msg = error_msg or str(e)
            logger.warning("HTTP Error: status=%s msg=%s", error_code, error_msg)
        except httpx.ConnectError as e:
            error_code = 503
            error_msg = error_msg or str(e)
            logger.warning("Connection Error: (%s) %s", error_code, error_msg)
        except httpx.TimeoutException:
            error_code = 408
            error_msg = error_msg or "Request timed out"
            logger.warning("Timeout Error: (%s) %s", error_code, error_msg)
        except Exception as e:
            error_code = error_code or 520
            error_msg = error_msg or str(e)
            logger.exception("Unexpected error: (%s) %s", error_code, error_msg)

        completed_at = time.monotonic()
        success = error_msg is None and error_code is None

        inter_chunk_times: List[float] = []
        num_delta_prompt_tokens = 0
        num_total_prompt_tokens = 0
        num_output_tokens = 0
        if success:
            end_to_end = max(0.0, completed_at - start_time)
            inter_chunk_times = [end_to_end]
            num_total_prompt_tokens = len(
                self.text_tokenizer_handle.encode(prompt_text)
            )
            num_delta_prompt_tokens = num_total_prompt_tokens

            completion_text = generated_text
            if (
                body.get("echo")
                and prompt_text
                and generated_text.startswith(prompt_text)
            ):
                completion_text = generated_text[len(prompt_text) :]

            num_output_tokens = len(self.text_tokenizer_handle.encode(completion_text))
            if max_tokens_limit is not None:
                num_output_tokens = min(int(max_tokens_limit), num_output_tokens)

        channels: Dict[ChannelModality, ChannelResponse] = {}
        if success:
            channels[ChannelModality.TEXT] = ChannelResponse(
                modality=ChannelModality.TEXT,
                content=completion_text,
                metrics={
                    "is_stream": False,
                    "inter_chunk_times": inter_chunk_times,
                    "num_delta_prompt_tokens": num_delta_prompt_tokens,
                    "num_total_prompt_tokens": num_total_prompt_tokens,
                    "num_output_tokens": num_output_tokens,
                    "logprobs": logprobs,
                },
            )

        return RequestResult(
            request_id=request.id,
            session_id=session_id,
            session_total_requests=session_total_requests,
            channels=channels,
            success=success,
            error_code=error_code,
            error_msg=error_msg,
            client_completed_at=completed_at,
        )



"""Unit tests for capacity search logic."""

from dataclasses import replace
from unittest.mock import MagicMock, patch

import pytest

from veeksha.capacity_search import (
    _adaptive_capacity_search,
    _knob_description,
    patch_traffic_knob,
    run_capacity_search,
)
from veeksha.config.benchmark import BenchmarkConfig
from veeksha.config.capacity_search import CapacitySearchConfig
from veeksha.config.generator.interval import FixedIntervalGeneratorConfig
from veeksha.config.traffic import ConcurrentTrafficConfig, RateTrafficConfig


@pytest.mark.unit
def test_adaptive_search_binary_logic_integer() -> None:
    """Test the adaptive search logic with integer domain (e.g. concurrency)."""
    
    # Passing: x <= 10. Max: 20. Start: 2. Expansion: 2.
    # Phase 1: 2(P) -> 4(P) -> 8(P) -> 16(F). Last passing=8, First failing=16.
    # Phase 2: Binary search in [8, 15].
    # Mid 12(F). Range [8, 11].
    # Mid 10(P). Best=10. Range [10, 11].
    # Mid 11(F). Range [10, 10].
    # Result: 10.
    
    def is_passing(x: float) -> bool:
        return x <= 10
        
    best, iters = _adaptive_capacity_search(
        start_value=2,
        max_value=20,
        expansion_factor=2.0,
        is_passing=is_passing,
        max_iterations=20,
        precision=0,
        integer_domain=True
    )
    
    assert best == 10
    assert iters < 20


@pytest.mark.unit
def test_adaptive_search_binary_logic_float() -> None:
    """Test the adaptive search logic with float domain (e.g. rate)."""
    
    # Passing: x <= 5.5. Start: 1.0. Max 10.0. Expansion 2.0.
    # Phase 1: 1.0(P) -> 2.0(P) -> 4.0(P) -> 8.0(F). Last=4.0, First=8.0.
    # Phase 2: Binary search [4.0, 8.0] precision 1 (integers 40 to 80).
    # 60 (6.0) -> Fail -> [40, 59]
    # 50 (5.0) -> Pass -> [50, 59], Best=5.0
    # 55 (5.5) -> Pass -> [55, 59], Best=5.5
    # 58 (5.8) -> Fail -> [55, 57]
    # 56 (5.6) -> Fail -> [55, 55]
    # Result -> 5.5
    
    def is_passing(x: float) -> bool:
        return x <= 5.55 # slightly lenient for float comparison, effectively testing 5.5
    
    best, iters = _adaptive_capacity_search(
        start_value=1.0,
        max_value=10.0,
        expansion_factor=2.0,
        is_passing=is_passing,
        max_iterations=20,
        precision=1,
        integer_domain=False
    )
    
    assert best == 5.5


@pytest.mark.unit
def test_patch_traffic_knob_rate() -> None:
    """Test patching rate traffic config."""
    # Setup benchmark config with RateTraffic
    interval = FixedIntervalGeneratorConfig(interval=1.0) # 1.0s -> 1.0 rps
    traffic = RateTrafficConfig(interval_generator=interval)
    config = BenchmarkConfig(traffic_scheduler=traffic)
    
    # Patch to 10.0 rps -> interval 0.1s
    new_config = patch_traffic_knob(config, value=10.0)
    
    new_traffic = new_config.traffic_scheduler
    assert isinstance(new_traffic, RateTrafficConfig)
    assert isinstance(new_traffic.interval_generator, FixedIntervalGeneratorConfig)
    assert new_traffic.interval_generator.interval == pytest.approx(0.1)


@pytest.mark.unit
def test_patch_traffic_knob_concurrent() -> None:
    """Test patching concurrent traffic config."""
    traffic = ConcurrentTrafficConfig(target_concurrent_sessions=1)
    config = BenchmarkConfig(traffic_scheduler=traffic)
    
    # Patch to 5.0
    new_config = patch_traffic_knob(config, value=5.0)
    
    new_traffic = new_config.traffic_scheduler
    assert isinstance(new_traffic, ConcurrentTrafficConfig)
    assert new_traffic.target_concurrent_sessions == 5
    assert new_traffic.rampup_seconds == 5 # Should auto-set rampup
    
    # Patch with non-integer -> ValueError
    with pytest.raises(ValueError):
        patch_traffic_knob(config, value=5.5)

@pytest.mark.unit
@patch("veeksha.capacity_search._read_slo_results")
@patch("veeksha.capacity_search.manage_benchmark_run")
@patch("veeksha.capacity_search._init_capacity_search_output_dir")
def test_run_capacity_search_e2e_flow(
    mock_init_dir,
    mock_manage_run,
    mock_read_slo,
    tmp_path
) -> None:
    """Test the full flow of run_capacity_search logic with mocks."""
    
    # Config
    start_value = 1
    max_value = 5
    
    # Passing criteria: <= 3 is pass, > 3 is fail.
    # 1(P) -> 2(P) -> 4(F). Binary [2, 3]. 3(P). Result 3.
    
    def side_effect_manage_run(cfg):
        # The value passed to patch_traffic_knob is embedded in cfg
        # But run_capacity_search calls patch_traffic_knob internally using the 'value' passed to run_one
        pass

    def side_effect_read_slo(output_dir):
        # We need to know what value was just run.
        # We can look at the latest call to manage_benchmark_run
        # The argument to manage_benchmark_run is the patched config.
        # We can inspect its traffic scheduler.
        
        args, _ = mock_manage_run.call_args
        run_cfg = args[0]
        val = run_cfg.traffic_scheduler.target_concurrent_sessions
        
        # 1(P) -> 2(P) -> 4(F). Binary [2, 3]. Mid 3(P). Result 3.
        if val <= 3:
            return {"all_slos_met": True}
        else:
            return {"all_slos_met": False}
    
    mock_read_slo.side_effect = side_effect_read_slo
    mock_init_dir.return_value = str(tmp_path)
    
    target_traffic = ConcurrentTrafficConfig(target_concurrent_sessions=1)
    base_config = BenchmarkConfig(traffic_scheduler=target_traffic)
    cap_config = CapacitySearchConfig(
        benchmark_config=base_config,
        start_value=start_value,
        max_value=max_value,
        expansion_factor=2.0,
        output_dir=str(tmp_path)
    )
    
    result = run_capacity_search(cap_config)
    
    assert result["best_value"] == 3
    assert result["max_iterations"] == 20
    assert len(result["history"]) == 5



from veeksha.config.generator.length import (
    FixedLengthGeneratorConfig,
)
from veeksha.generator.length.base import (
    BaseLengthGenerator,
)


class FixedLengthGenerator(BaseLengthGenerator):
    def __init__(self, config: FixedLengthGeneratorConfig, rng=None):
        self.config = config
        self.rng = rng

    def get_next_value(self) -> int:
        return self.config.value



from typing import List

from veeksha.config.server import VajraServerConfig
from veeksha.logger import init_logger
from veeksha.orchestration.server_manager import BaseServerManager

logger = init_logger(__name__)


class VajraServerManager(BaseServerManager):
    """Manager for Vajra inference servers."""

    def __init__(self, config: VajraServerConfig, output_dir: str):
        super().__init__(config, output_dir=output_dir)

        if config.engine.lower() != "vajra":
            logger.warning(
                f"VajraServerManager created with engine='{config.engine}'. "
                "Expected 'vajra'"
            )

    def _build_launch_command(self) -> List[str]:
        # Placeholder for Vajra command as it was not fully implemented in previous steps
        # Assuming similar structure
        command = [
            "python",
            "-m",
            "vajra_server.server",
            "--model",
            self.config.model,
            "--host",
            self.config.host,
            "--port",
            str(self.config.port),
        ]

        if self.config.tensor_parallel_size > 1:
            command.extend(
                ["--tensor-parallel-size", str(self.config.tensor_parallel_size)]
            )

        if self.config.api_key:
            command.extend(["--api-key", self.config.api_key])

        if self.config.max_model_len:
            command.extend(["--max-model-len", str(self.config.max_model_len)])

        additional_args_dict = self.get_additional_args_dict()
        for key, value in additional_args_dict.items():
            if isinstance(value, bool):
                if value:
                    command.append(f"--{key}")
            elif isinstance(value, list):
                for item in value:
                    command.extend([f"--{key}", str(item)])
            elif value is not None:
                command.extend([f"--{key}", str(value)])

        return command



"""LLM clients for the new Veeksha framework."""

from veeksha.client.base import BaseLLMClient
from veeksha.client.openai_chat import OpenAIChatCompletionsClient
from veeksha.client.openai_router import OpenAIRouterClient
from veeksha.client.registry import ClientRegistry

__all__ = [
    "BaseLLMClient",
    "OpenAIChatCompletionsClient",
    "OpenAIRouterClient",
    "ClientRegistry",
]



import math

import numpy as np

from veeksha.config.generator.interval import (
    PoissonIntervalGeneratorConfig,
)
from veeksha.generator.interval.base import BaseIntervalGenerator


class PoissonIntervalGenerator(BaseIntervalGenerator):
    def __init__(
        self,
        config: PoissonIntervalGeneratorConfig,
        rng: np.random.RandomState,
    ):
        self.config = config
        self.rng = rng

        self.arrival_rate = self.config.arrival_rate
        self.std = 1.0 / self.arrival_rate
        self.max_interval = self.std * 3.0

    def get_next_interval(self) -> float:
        next_interval = (
            -math.log(1.0 - float(self.rng.random_sample())) / self.arrival_rate
        )
        next_interval = min(next_interval, self.max_interval)
        return next_interval



"""Prefetch worker for session generation and scheduling."""

import threading
import time
from typing import Optional

from veeksha.core.context import WorkerContext
from veeksha.core.session import Session
from veeksha.generator.session.base import BaseSessionGenerator
from veeksha.logger import init_logger
from veeksha.traffic.base import BaseTrafficScheduler

logger = init_logger(__name__)


class SharedSessionCounter:
    """Thread-safe shared counter for tracking sessions across workers."""

    def __init__(self, max_sessions: int = -1):
        self.max_sessions = max_sessions
        self._count = 0

    def try_increment(self) -> bool:
        if self.max_sessions < 0:
            self._count += 1
            return True
        if self._count < self.max_sessions:
            self._count += 1
            return True
        return False

    @property
    def count(self) -> int:
        return self._count


class PrefetchWorker:
    """Worker that generates sessions and schedules them with the traffic scheduler.

    This worker pulls sessions from the session generator and feeds them to the
    traffic scheduler, which then manages the dispatch timing of individual requests.
    """

    # unthrottled for first 3 seconds, then throttles
    _BURST_DURATION_S = 5.0
    _MAX_POLL_INTERVAL_S = 0.05

    def __init__(
        self,
        traffic_scheduler: BaseTrafficScheduler,
        session_generator: BaseSessionGenerator,
        generator_lock: threading.Lock,
        worker_context: WorkerContext,
        session_counter: SharedSessionCounter,
    ):
        """Initialize the prefetch worker.

        Args:
            traffic_scheduler: Scheduler to schedule sessions with
            session_generator: Generator to get sessions from
            generator_lock: Lock protecting the session generator
            worker_context: Worker context with stop event
            session_counter: Shared counter for tracking sessions across workers
        """
        self.traffic_scheduler = traffic_scheduler
        self.session_generator = session_generator
        self.generator_lock = generator_lock
        self.worker_context = worker_context
        self.session_counter = session_counter

    def _get_poll_interval(self) -> float:
        """Calculate poll interval based on runtime duration.

        Unthrottled for the first _BURST_DURATION_S seconds, then throttles
        to _MAX_POLL_INTERVAL_S.

        Returns:
            Poll interval in seconds.
        """
        if time.monotonic() - self._start_time < self._BURST_DURATION_S:
            return 0.0
        return self._MAX_POLL_INTERVAL_S

    def _generate_session(self) -> Optional[Session]:
        """Generate next session in a thread-safe manner."""
        while not self.worker_context.stop_event.is_set():
            with self.generator_lock:
                if not self.session_counter.try_increment():
                    return None  # exhausted

                try:
                    session = self.session_generator.generate_session()
                    return session
                except StopIteration:
                    logger.debug(
                        "Prefetch worker %s: generator exhausted",
                        self.worker_context.worker_id,
                    )
                    return None

        return None

    def run(self) -> None:
        """Main worker loop."""
        logger.debug("Prefetch worker %s starting", self.worker_context.worker_id)

        self._start_time = time.monotonic()

        while not self.worker_context.stop_event.is_set():
            session = self._generate_session()
            if session is None:
                logger.info(
                    "Prefetch worker %s: no more sessions to generate",
                    self.worker_context.worker_id,
                )
                break

            # Schedule the session with traffic scheduler
            self.traffic_scheduler.schedule_session(session)

            if self.session_counter.count % 100 == 0:
                logger.debug(
                    "Prefetch progress: %d sessions generated",
                    self.session_counter.count,
                )

            # Throttle (burst at start, then steady-state)
            time.sleep(self._get_poll_interval())

        logger.debug("Prefetch worker %s exiting", self.worker_context.worker_id)



from abc import abstractmethod
from typing import Mapping, Optional, Set, Tuple

from veeksha.config.traffic import BaseTrafficConfig
from veeksha.core.request import Request
from veeksha.core.response import ChannelResponse
from veeksha.core.seeding import SeedManager
from veeksha.core.session import Session
from veeksha.types import ChannelModality


class BaseTrafficScheduler:
    def __init__(self, config: BaseTrafficConfig, seed_manager: SeedManager):
        self.config = config
        self.seed_manager = seed_manager

    @abstractmethod
    def schedule_session(self, session: Session) -> None:
        """Schedule a session for dispatch."""
        raise NotImplementedError

    @abstractmethod
    def pop_ready(self) -> Optional[Tuple[Request, int, int]]:
        """Pop a ready request from the scheduler.

        Returns:
            Tuple of (request, session_id, session_size) if a request is ready,
            None otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def wait_for_ready(
        self, timeout: float = 0.001
    ) -> Optional[Tuple[Request, int, int]]:
        """Wait for a ready request with timeout.

        Args:
            timeout: Maximum time to wait in seconds.

        Returns:
            Tuple of (request, session_id, session_size) if ready, None if timeout.
        """
        raise NotImplementedError

    @abstractmethod
    def notify_completion(
        self,
        request_id: int,
        completed_at_monotonic: float,
        success: bool,
        channel_responses: Optional[Mapping[ChannelModality, ChannelResponse]] = None,
    ) -> None:
        """Notify the scheduler that a request has completed."""
        raise NotImplementedError

    @abstractmethod
    def get_session_id(self, request_id: int) -> int:
        """Get the session ID for a given request ID.

        Returns -1 if the request is not found.
        """
        raise NotImplementedError

    @abstractmethod
    def get_session_size(self, request_id: int) -> int:
        """Get the total number of requests in the session for a given request ID.

        Returns 1 if the request is not found.
        """
        raise NotImplementedError

    @abstractmethod
    def has_pending_work(self) -> bool:
        """Check if there are pending sessions or in-flight requests."""
        raise NotImplementedError

    @abstractmethod
    def get_in_flight_request_ids(self) -> Set[int]:
        """Return the set of request IDs currently in-flight."""
        raise NotImplementedError

    def reset_reference_time(self) -> None:
        """Optional hook invoked before the benchmark starts dispatching."""
        return



from __future__ import annotations

from veeksha.config.generator.length import StairLengthGeneratorConfig
from veeksha.generator.length.base import BaseLengthGenerator


class StairLengthGenerator(BaseLengthGenerator):
    """Deterministic stair-step generator.

    Emits the configured values in-order, repeating each step `repeat_each` times.
    When `wrap` is enabled, cycles back to the beginning after the last step.
    """

    def __init__(self, config: StairLengthGeneratorConfig, rng=None):
        self.config = config
        self.rng = rng

        self._step_idx = 0
        self._repeat_idx = 0

    def get_next_value(self) -> int:
        value = self.config.values[self._step_idx]

        self._repeat_idx += 1
        if self._repeat_idx >= self.config.repeat_each:
            self._repeat_idx = 0
            self._step_idx += 1
            if self._step_idx >= len(self.config.values):
                self._step_idx = 0 if self.config.wrap else len(self.config.values) - 1

        return int(value)




import pytest
import numpy as np
from veeksha.config.generator.length import (
    FixedLengthGeneratorConfig,
    UniformLengthGeneratorConfig,
    ZipfLengthGeneratorConfig,
)
from veeksha.generator.length.fixed import FixedLengthGenerator
from veeksha.generator.length.uniform import UniformLengthGenerator
from veeksha.generator.length.zipf import ZipfLengthGenerator

@pytest.fixture
def rng():
    return np.random.RandomState(42)

def test_fixed_length_generator():
    config = FixedLengthGeneratorConfig(value=10)
    generator = FixedLengthGenerator(config)
    assert generator.get_next_value() == 10

def test_uniform_length_generator(rng):
    config = UniformLengthGeneratorConfig(min=5, max=10)
    generator = UniformLengthGenerator(config, rng)
    
    values = [generator.get_next_value() for _ in range(100)]
    assert all(5 <= v < 10 for v in values) # numpy uniform is [low, high) for floats, but let's check exact implementation
    # Implementation: int(rng.uniform(min, max)). So [min, max).
    # Wait, python's random.uniform includes max, but numpy.random.uniform excludes max.
    # The implementation uses self.rng.uniform(self.config.min, self.config.max).
    # We should verify if max is inclusive or exclusive based on implementation details or desired behavior.
    # Assuming standard numpy behavior, it's exclusive of max.
    
    assert min(values) >= 5
    assert max(values) < 10

def test_zipf_length_generator(rng):
    config = ZipfLengthGeneratorConfig(min=1, max=100, theta=1.5, scramble=False)
    generator = ZipfLengthGenerator(config, rng)
    
    values = [generator.get_next_value() for _ in range(100)]
    assert all(1 <= v <= 100 for v in values)
    # Zipf distribution creates many small values
    assert np.mean(values) < 50

def test_zipf_length_generator_scramble(rng):
    config = ZipfLengthGeneratorConfig(min=1, max=100, theta=1.5, scramble=True)
    generator = ZipfLengthGenerator(config, rng)
    
    values = [generator.get_next_value() for _ in range(100)]
    assert all(1 <= v <= 100 for v in values)



from abc import ABC
from typing import Any

from veeksha.config.core.frozen_dataclass import frozen_dataclass
from veeksha.config.utils import get_all_subclasses


@frozen_dataclass
class BasePolyConfig(ABC):
    @classmethod
    def create_from_type(cls, type_: Any) -> Any:
        for subclass in get_all_subclasses(cls):
            if subclass.get_type() == type_:
                return subclass()
        raise ValueError(f"Invalid type: {type_}")

    @classmethod
    def get_type(cls) -> Any:
        raise NotImplementedError(
            f"[{cls.__name__}] get_type() method is not implemented"
        )



import pytest

from veeksha.config.evaluator import PerformanceEvaluatorConfig
from veeksha.core.request_content import TextChannelRequestContent
from veeksha.core.response import ChannelResponse, RequestResult
from veeksha.evaluator.performance.text import TextPerformanceEvaluator
from veeksha.types import ChannelModality


@pytest.mark.unit
def test_text_performance_skips_tpot_for_non_stream_request() -> None:
    config = PerformanceEvaluatorConfig(stream_metrics=False)
    evaluator = TextPerformanceEvaluator(
        config=config,
        channel_config=config.text_channel,
        benchmark_start_time=0.0,
    )

    evaluator.register_request(
        request_id=0,
        session_id=0,
        dispatched_at=0.0,
        content=TextChannelRequestContent(
            input_text="hello",
            target_output_tokens=1,
            target_prompt_tokens=1,
        ),
    )

    result = RequestResult(
        request_id=0,
        session_id=0,
        session_total_requests=1,
        channels={
            ChannelModality.TEXT: ChannelResponse(
                modality=ChannelModality.TEXT,
                content="x",
                metrics={
                        "is_stream": False,
                    # Non-stream: only one duration for the whole response
                    "inter_chunk_times": [0.5],
                    "num_delta_prompt_tokens": 1,
                    "num_total_prompt_tokens": 1,
                    "num_output_tokens": 1,
                },
            )
        },
        success=True,
        client_completed_at=0.5,
    )

    evaluator.record_request_completed(
        request_id=0,
        session_id=0,
        completed_at=0.5,
        response=result,
    )

    assert len(evaluator.summaries["ttfc"]) == 1
    assert len(evaluator.summaries["tpot"]) == 0


@pytest.mark.unit
def test_text_performance_includes_tpot_for_streaming_request() -> None:
    config = PerformanceEvaluatorConfig(stream_metrics=False)
    evaluator = TextPerformanceEvaluator(
        config=config,
        channel_config=config.text_channel,
        benchmark_start_time=0.0,
    )

    evaluator.register_request(
        request_id=0,
        session_id=0,
        dispatched_at=0.0,
        content=TextChannelRequestContent(
            input_text="hello",
            target_output_tokens=2,
            target_prompt_tokens=1,
        ),
    )

    result = RequestResult(
        request_id=0,
        session_id=0,
        session_total_requests=1,
        channels={
            ChannelModality.TEXT: ChannelResponse(
                modality=ChannelModality.TEXT,
                content="xy",
                metrics={
                    "is_stream": True,
                    # Streaming: at least TTFC + one chunk gap
                    "inter_chunk_times": [0.1, 0.2],
                    "num_delta_prompt_tokens": 1,
                    "num_total_prompt_tokens": 1,
                    "num_output_tokens": 2,
                },
            )
        },
        success=True,
        client_completed_at=0.3,
    )

    evaluator.record_request_completed(
        request_id=0,
        session_id=0,
        completed_at=0.3,
        response=result,
    )

    assert len(evaluator.summaries["tpot"]) == 1





from __future__ import annotations

import math

import pytest

from veeksha.evaluator.accuracy.base import LMEvalAccuracyEvaluator


def _parse(ctxlen: int, lp: dict) -> tuple[float, bool]:
    # Instantiate without calling __init__ to avoid heavyweight config wiring.
    ev = LMEvalAccuracyEvaluator.__new__(LMEvalAccuracyEvaluator)
    return LMEvalAccuracyEvaluator._parse_logprobs(ev, ctxlen=ctxlen, lp=lp)


def test_parse_logprobs_openai_dict_sums_target_and_ignores_last_completion() -> None:
    # token_logprobs includes ctx (2 toks), target (2 toks), completion (1 tok).
    lp = {
        "token_logprobs": [-10.0, -9.0, -1.5, -2.5, -99.0],
        "top_logprobs": [
            {"a": -10.0},
            {"b": -9.0},
            {"T1": -1.5, "x": -2.0},
            {"T2": -2.5, "y": -3.0},
            {"C": -99.0, "z": -100.0},
        ],
    }
    total, is_greedy = _parse(ctxlen=2, lp=lp)
    assert total == pytest.approx(-4.0)
    assert is_greedy is True


def test_parse_logprobs_openai_dict_marks_non_greedy_when_top_beats_token() -> None:
    lp = {
        "token_logprobs": [-1.0, -2.0, -3.0],
        "top_logprobs": [
            {"a": -1.0},
            {"b": -1.9, "other": -1.0},  # token is not argmax
            {"c": -3.0},
        ],
    }
    total, is_greedy = _parse(ctxlen=0, lp=lp)
    # Slicing ctxlen:-1 ignores the last entry, so sum is over first two.
    assert total == pytest.approx(-3.0)
    assert is_greedy is False


def test_parse_logprobs_content_list() -> None:
    lp = {
        "content": [
            {"token": "c1", "logprob": -9.0, "top_logprobs": [{"token": "c1", "logprob": -9.0}]},
            {"token": "c2", "logprob": -8.0, "top_logprobs": [{"token": "c2", "logprob": -8.0}]},
            {"token": "t1", "logprob": -1.0, "top_logprobs": [{"token": "t1", "logprob": -1.0}]},
            {"token": "t2", "logprob": -2.0, "top_logprobs": [{"token": "t2", "logprob": -2.0}]},
        ]
    }
    total, is_greedy = _parse(ctxlen=2, lp=lp)
    assert total == pytest.approx(-3.0)
    assert is_greedy is True


def test_parse_logprobs_chunks_list_token_logprobs_fallback() -> None:
    lp = {
        "chunks": [
            {"token_logprobs": [-1.0], "top_logprobs": [{"token": "a", "logprob": -1.0}]},
            {"token_logprobs": [-2.0], "top_logprobs": [{"token": "b", "logprob": -2.0}]},
        ]
    }
    total, is_greedy = _parse(ctxlen=123, lp=lp)
    assert total == pytest.approx(-3.0)
    assert is_greedy is True


def test_parse_logprobs_raises_on_unknown_format() -> None:
    with pytest.raises(KeyError):
        _parse(ctxlen=0, lp={"weird": 1})





import unittest
from veeksha.config.slo import ConstantSloConfig
from veeksha.slo.runner import _get_slo_metric_keys

class TestSloRunner(unittest.TestCase):
    def test_get_slo_metric_keys_mapping(self):
        configs = [
            ConstantSloConfig(metric="e2e", value=1.0),
            ConstantSloConfig(metric="tpot", value=0.1),
            ConstantSloConfig(metric="ttfc", value=0.5),
        ]
        keys = _get_slo_metric_keys(configs)
        
        # e2e should map to end_to_end_latency
        self.assertIn("end_to_end_latency", keys)
        # tpot should be present
        self.assertIn("tpot", keys)
        # ttfc should be present
        self.assertIn("ttfc", keys)

if __name__ == "__main__":
    unittest.main()



from veeksha.types import LengthGeneratorType
from veeksha.types.base_registry import BaseRegistry

from .fixed import FixedLengthGenerator
from .stair import StairLengthGenerator
from .uniform import UniformLengthGenerator
from .zipf import ZipfLengthGenerator


class LengthGeneratorRegistry(BaseRegistry):
    @classmethod
    def get_key_from_str(cls, key_str: str) -> LengthGeneratorType:
        return LengthGeneratorType.from_str(key_str)  # type: ignore


LengthGeneratorRegistry.register(LengthGeneratorType.ZIPF, ZipfLengthGenerator)
LengthGeneratorRegistry.register(LengthGeneratorType.UNIFORM, UniformLengthGenerator)
LengthGeneratorRegistry.register(LengthGeneratorType.FIXED, FixedLengthGenerator)
LengthGeneratorRegistry.register(LengthGeneratorType.FIXED_STAIR, StairLengthGenerator)



import pytest
from veeksha.config.generator.session_graph import LinearSessionGraphGeneratorConfig
from veeksha.config.generator.length import UniformLengthGeneratorConfig
from veeksha.core.seeding import SeedManager
from veeksha.generator.session_graph.linear import LinearSessionGraphGenerator
from veeksha.core.session_graph import SessionGraph


@pytest.fixture
def seed_manager():
    return SeedManager(seed=42)


@pytest.mark.unit
def test_linear_graph_generator_inherit_history_true(seed_manager):
    """Verify that edges have is_history_parent=True when configured."""
    # Configure generator to produce exactly 2 requests
    config = LinearSessionGraphGeneratorConfig(
        num_request_generator=UniformLengthGeneratorConfig(min=2, max=2),
        inherit_history=True
    )
    generator = LinearSessionGraphGenerator(config, seed_manager)
    graph = generator.generate_session_graph()
    
    # Check edges
    # Graph should have 2 nodes (0, 1) and 1 edge (0->1)
    assert len(graph.nodes) == 2
    edges = []
    for edge_list in graph.outgoing.values():
        edges.extend(edge_list)
    assert len(edges) == 1
    
    edge = edges[0]
    assert edge.src == 0
    assert edge.dst == 1
    assert edge.is_history_parent is True


@pytest.mark.unit
def test_linear_graph_generator_inherit_history_false(seed_manager):
    """Verify that edges have is_history_parent=False when configured."""
    # Configure generator to produce exactly 2 requests
    config = LinearSessionGraphGeneratorConfig(
        num_request_generator=UniformLengthGeneratorConfig(min=2, max=2),
        inherit_history=False
    )
    generator = LinearSessionGraphGenerator(config, seed_manager)
    graph = generator.generate_session_graph()
    
    # Check edges
    # Graph should have 2 nodes (0, 1) and 1 edge (0->1)
    assert len(graph.nodes) == 2
    edges = []
    for edge_list in graph.outgoing.values():
        edges.extend(edge_list)
    assert len(edges) == 1
    
    edge = edges[0]
    assert edge.src == 0
    assert edge.dst == 1
    assert edge.is_history_parent is False



from __future__ import annotations

"""Plotting utilities for deciding and applying axis scale transforms.

This module provides a minimal, reusable heuristic to choose a numeric axis scale
and small helpers to transform values and format labels accordingly.
"""

from typing import Iterable, Literal

import numpy as np

AxisScale = Literal["linear", "log", "symlog"]
AxisChoice = Literal["x", "y"]


def _finite_array(values: Iterable[float]) -> np.ndarray:
    arr = np.asarray(list(values), dtype=float)
    if arr.size == 0:
        return arr
    mask = np.isfinite(arr)
    return arr[mask]


def recommend_axis_scale(
    values: Iterable[float],
    *,
    kind: Literal["numeric", "probability"] = "numeric",
    threshold: float = 50.0,
    min_samples: int = 10,
) -> AxisScale:
    """Recommend an axis scale for numeric plotting.

    Args:
        values: Sequence of numeric values to be plotted along the axis.
        kind: Treat data as 'numeric' or 'probability'. Probability is always linear.
        threshold: Robust dynamic range (p95/p5) needed to switch to log/symlog.
        min_samples: Minimum number of finite samples required to make a decision.

    Returns:
        One of 'linear', 'log', 'symlog'.
    """
    arr = _finite_array(values)
    if arr.size < min_samples:
        return "linear"

    if kind == "probability":
        return "linear"

    # If all values are within [0, 1], assume probability-like and keep linear.
    if arr.size > 0 and np.nanmin(arr) >= 0.0 and np.nanmax(arr) <= 1.0:
        return "linear"

    has_non_positive = np.any(arr <= 0)
    positives = arr[arr > 0]

    if positives.size == 0:
        # No positive mass; only symlog can reasonably represent wide ranges.
        return "symlog"

    q5, q95 = np.quantile(positives, [0.05, 0.95])
    q5 = max(q5, 1e-12)
    robust_ratio = float(q95 / q5) if q5 > 0 else float("inf")

    if robust_ratio >= threshold:
        return "symlog" if has_non_positive else "log"

    return "linear"


def format_axis_label(base: str, unit: str | None, scale: AxisScale) -> str:
    """Format axis label with unit and scale indicator.

    Example:
        format_axis_label("TTFT", "s", "log") -> "TTFT (s, log scale)"
    """
    if scale == "linear":
        return f"{base}{f' ({unit})' if unit else ''}"
    suffix = "log scale" if scale == "log" else "symlog"
    if unit:
        return f"{base} ({unit}, {suffix})"
    return f"{base} ({suffix})"


def apply_axis_scale(
    fig,
    axis: AxisChoice,
    scale: AxisScale,
    *,
    symlog_linthresh: float | None = None,
) -> None:
    """Attempt to set matplotlib axis scale on a Rekha figure.

    Best-effort: tries common attributes to access the underlying matplotlib Axes.
    """
    if scale == "linear":
        return
    try:
        # Preferred: Rekha exposes underlying matplotlib axes via get_axes()
        ax = fig.get_axes()[0]

        if scale == "log":
            if axis == "x":
                ax.set_xscale("log")
            else:
                ax.set_yscale("log")
            return

        # symlog
        linthresh = 1e-6 if symlog_linthresh is None else float(symlog_linthresh)
        if axis == "x":
            ax.set_xscale("symlog", linthresh=linthresh)
        else:
            ax.set_yscale("symlog", linthresh=linthresh)
    except Exception:
        # plots will remain linear.
        return



"""Unit tests for RateTrafficScheduler."""

import time
from typing import Any, Dict, Optional

import pytest

from veeksha.config.generator.interval import FixedIntervalGeneratorConfig
from veeksha.config.traffic import RateTrafficConfig
from veeksha.core.request import Request
from veeksha.core.request_content import TextChannelRequestContent
from veeksha.core.response import ChannelResponse
from veeksha.core.seeding import SeedManager
from veeksha.core.session import Session
from veeksha.core.session_graph import SessionEdge, SessionGraph, SessionNode, add_edge, add_node
from veeksha.traffic.rate import RateTrafficScheduler
from veeksha.types import ChannelModality


def wait_until(predicate, timeout_s=0.5, interval_s=0.005):
    """Wait until predicate returns True or timeout."""
    start = time.monotonic()
    while time.monotonic() - start < timeout_s:
        if predicate():
            return True
        time.sleep(interval_s)
    return False


def pop_ready_with_timeout(scheduler, timeout_s=0.2) -> Optional[Request]:
    """Helper to pop a ready request with a timeout."""
    start = time.monotonic()
    while time.monotonic() - start < timeout_s:
        result = scheduler.pop_ready()
        if result is not None:
            return result[0]
        time.sleep(0.005)
    return None


def make_request(request_id: int, model: str = "dummy") -> Request:
    """Create a simple test request."""
    return Request(
        id=request_id,
        channels={
            ChannelModality.TEXT: TextChannelRequestContent(
                input_text=f"test_{request_id}", target_output_tokens=10
            )
        },
    )


def make_linear_session(session_id: int, num_requests: int, wait_times: Optional[list] = None) -> Session:
    """Create a session with a linear chain of requests: 0 -> 1 -> 2 -> ..."""
    graph = SessionGraph()
    requests: Dict[int, Request] = {}
    
    if wait_times is None:
        wait_times = [0.0] * num_requests
    
    for i in range(num_requests):
        add_node(graph, SessionNode(id=i, wait_after_ready=wait_times[i]))
        requests[i] = make_request(request_id=session_id * 100 + i)
    
    for i in range(num_requests - 1):
        add_edge(graph, SessionEdge(src=i, dst=i + 1))
    
    return Session(id=session_id, session_graph=graph, requests=requests)


def make_scheduler(interval: float = 0.1) -> RateTrafficScheduler:
    """Create a scheduler with fixed interval generator."""
    config = RateTrafficConfig(
        interval_generator=FixedIntervalGeneratorConfig(interval=interval)
    )
    seed_manager = SeedManager(seed=42)
    return RateTrafficScheduler(config, seed_manager)


@pytest.mark.unit
def test_single_request_session_dispatches() -> None:
    """Single-request session dispatches at the generated start time."""
    scheduler = make_scheduler(interval=0.05)
    session = make_linear_session(session_id=1, num_requests=1)
    
    scheduler.schedule_session(session)
    
    # Should not be ready immediately (first session starts at t=0, but need tiny delay for processing)
    # With interval=0.05, first session starts at t=0
    assert wait_until(lambda: scheduler.pop_ready() is not None, timeout_s=0.2)


@pytest.mark.unit
def test_linear_chain_respects_dependencies() -> None:
    """Linear chain waits for each request to complete before releasing next."""
    scheduler = make_scheduler(interval=0.01)
    session = make_linear_session(session_id=2, num_requests=3, wait_times=[0.0, 0.02, 0.02])
    
    scheduler.schedule_session(session)
    
    # First request ready
    req1 = None
    assert wait_until(lambda: (req1 := scheduler.pop_ready()) is not None, timeout_s=0.2)
    
    # Second not ready yet (first not completed)
    assert scheduler.pop_ready() is None
    
    # Complete first request
    scheduler.notify_completion(request_id=200, completed_at_monotonic=time.monotonic(), success=True)
    
    # Second becomes ready after its wait_after_ready
    assert wait_until(lambda: scheduler.pop_ready() is not None, timeout_s=0.2)


@pytest.mark.unit
def test_cancellation_drops_pending_requests() -> None:
    """Failure with cancel_on_failure drops remaining session requests."""
    scheduler = make_scheduler(interval=0.01)
    session = make_linear_session(session_id=3, num_requests=3)
    
    scheduler.schedule_session(session)
    
    # First request ready
    assert wait_until(lambda: scheduler.pop_ready() is not None, timeout_s=0.2)
    
    # Fail the first request
    scheduler.notify_completion(request_id=300, completed_at_monotonic=time.monotonic(), success=False)
    
    # No more requests should become ready
    time.sleep(0.05)
    assert scheduler.pop_ready() is None


@pytest.mark.unit
def test_multiple_sessions_interleave() -> None:
    """Multiple sessions interleave without interference."""
    scheduler = make_scheduler(interval=0.01)
    session1 = make_linear_session(session_id=10, num_requests=2)
    session2 = make_linear_session(session_id=20, num_requests=2)
    
    scheduler.schedule_session(session1)
    scheduler.schedule_session(session2)
    
    # Both first requests should become ready
    ready_count = 0
    for _ in range(10):
        if wait_until(lambda: scheduler.pop_ready() is not None, timeout_s=0.1):
            ready_count += 1
        if ready_count >= 2:
            break
    
    assert ready_count >= 2, "Expected both root requests to become ready"


@pytest.mark.unit
def test_fan_out_session() -> None:
    """Session with fan-out: one root spawns multiple parallel children."""
    scheduler = make_scheduler(interval=0.01)
    
    # Create graph: 0 -> 1, 0 -> 2 (fan-out)
    graph = SessionGraph()
    add_node(graph, SessionNode(id=0, wait_after_ready=0.0))
    add_node(graph, SessionNode(id=1, wait_after_ready=0.0))
    add_node(graph, SessionNode(id=2, wait_after_ready=0.0))
    add_edge(graph, SessionEdge(src=0, dst=1))
    add_edge(graph, SessionEdge(src=0, dst=2))
    
    requests = {
        0: make_request(400),
        1: make_request(401),
        2: make_request(402),
    }
    session = Session(id=40, session_graph=graph, requests=requests)
    
    scheduler.schedule_session(session)
    
    # Root (0) should be ready
    assert wait_until(lambda: scheduler.pop_ready() is not None, timeout_s=0.2)
    
    # Children not ready yet
    assert scheduler.pop_ready() is None
    
    # Complete root
    scheduler.notify_completion(request_id=400, completed_at_monotonic=time.monotonic(), success=True)
    
    # Both children should become ready
    child1 = None
    child2 = None
    assert wait_until(lambda: (child1 := scheduler.pop_ready()) is not None, timeout_s=0.2)
    assert wait_until(lambda: (child2 := scheduler.pop_ready()) is not None, timeout_s=0.2)


@pytest.mark.unit
def test_fan_in_session() -> None:
    """Session with fan-in: multiple parents must complete before child is ready."""
    scheduler = make_scheduler(interval=0.01)
    
    # Create graph: 0 -> 2, 1 -> 2 (fan-in to node 2)
    graph = SessionGraph()
    add_node(graph, SessionNode(id=0, wait_after_ready=0.0))
    add_node(graph, SessionNode(id=1, wait_after_ready=0.0))
    add_node(graph, SessionNode(id=2, wait_after_ready=0.0))
    add_edge(graph, SessionEdge(src=0, dst=2, is_history_parent=False))
    add_edge(graph, SessionEdge(src=1, dst=2, is_history_parent=True))
    
    requests = {
        0: make_request(500),
        1: make_request(501),
        2: make_request(502),
    }
    session = Session(id=50, session_graph=graph, requests=requests)
    
    scheduler.schedule_session(session)
    
    # Both roots should be ready (0 and 1 have no parents)
    root1 = None
    root2 = None
    assert wait_until(lambda: (root1 := scheduler.pop_ready()) is not None, timeout_s=0.2)
    assert wait_until(lambda: (root2 := scheduler.pop_ready()) is not None, timeout_s=0.2)
    
    # Child (2) not ready yet
    assert scheduler.pop_ready() is None
    
    # Complete only one parent
    scheduler.notify_completion(request_id=500, completed_at_monotonic=time.monotonic(), success=True)
    
    # Child still not ready (other parent not done)
    time.sleep(0.02)
    assert scheduler.pop_ready() is None
    
    # Complete second parent
    scheduler.notify_completion(request_id=501, completed_at_monotonic=time.monotonic(), success=True)
    
    # Now child should be ready
    assert wait_until(lambda: scheduler.pop_ready() is not None, timeout_s=0.2)


@pytest.mark.unit
def test_session_garbage_collected_when_complete() -> None:
    """Session state is cleaned up after all requests complete."""
    scheduler = make_scheduler(interval=0.01)
    session = make_linear_session(session_id=60, num_requests=1)
    
    scheduler.schedule_session(session)
    assert len(scheduler._sessions) == 1
    
    # Pop and complete the request
    assert wait_until(lambda: scheduler.pop_ready() is not None, timeout_s=0.2)
    scheduler.notify_completion(request_id=6000, completed_at_monotonic=time.monotonic(), success=True)
    
    # Session should be garbage collected
    assert len(scheduler._sessions) == 0


@pytest.mark.unit
def test_history_inheritance() -> None:
    """Requests inherit history from their history parent."""
    scheduler = make_scheduler(interval=0.01)
    
    # Linear: 0 -> 1 -> 2
    session = make_linear_session(session_id=1, num_requests=3)
    scheduler.schedule_session(session)
    
    # 0 ready
    req0 = pop_ready_with_timeout(scheduler)
    assert req0 is not None
    assert req0.history == []
    
    # Complete 0 with history
    response_0 = ChannelResponse(
        modality=ChannelModality.TEXT,
        content="response_0"
    )
    scheduler.notify_completion(
        request_id=req0.id, 
        completed_at_monotonic=time.monotonic(), 
        success=True, 
        channel_responses={ChannelModality.TEXT: response_0}
    )
    
    # 1 ready
    req1 = pop_ready_with_timeout(scheduler)
    assert req1 is not None
    assert len(req1.history) == 2
    assert req1.history[0] == {"role": "user", "content": "test_100"}
    assert req1.history[1] == {"role": "assistant", "content": "response_0"}
    
    # Complete 1 with history
    response_1 = ChannelResponse(
        modality=ChannelModality.TEXT,
        content="response_1"
    )
    scheduler.notify_completion(
        request_id=req1.id, 
        completed_at_monotonic=time.monotonic(), 
        success=True, 
        channel_responses={ChannelModality.TEXT: response_1}
    )
    
    # 2 ready
    req2 = pop_ready_with_timeout(scheduler)
    assert req2 is not None
    assert len(req2.history) == 4


@pytest.mark.unit
def test_ambiguous_history_inheritance() -> None:
    """Raises ValueError if multiple history parents exist."""
    scheduler = make_scheduler(interval=0.01)
    
    # Graph: 0 -> 2, 1 -> 2
    graph = SessionGraph()
    add_node(graph, SessionNode(id=0, wait_after_ready=0.0))
    add_node(graph, SessionNode(id=1, wait_after_ready=0.0))
    add_node(graph, SessionNode(id=2, wait_after_ready=0.0))
    
    # Both are history parents
    add_edge(graph, SessionEdge(src=0, dst=2, is_history_parent=True))
    add_edge(graph, SessionEdge(src=1, dst=2, is_history_parent=True))
    
    requests = {
        0: make_request(200),
        1: make_request(201),
        2: make_request(202),
    }
    session = Session(id=2, session_graph=graph, requests=requests)
    scheduler.schedule_session(session)
    
    # Complete parents (0 and 1)
    # We expect exactly 2 requests to be processed successfully
    for _ in range(2):
        req = pop_ready_with_timeout(scheduler)
        assert req is not None
        scheduler.notify_completion(req.id, time.monotonic(), success=True)
    
    # 2 should be ready now, but pop_ready checks history ambiguity
    # Wait for it to be in ready queue (internal) but popping fails
    # Since pop_ready raises, we can just check:
    def check_raises():
        try:
             req = scheduler.pop_ready()
             return False # Should have raised or returned None if not ready yet
        except ValueError:
             return True
    
    assert wait_until(check_raises, timeout_s=1.0)


@pytest.mark.unit
def test_no_history_parent() -> None:
    """Requests with no history parent start with empty history."""
    scheduler = make_scheduler(interval=0.01)
    
    # Graph: 0 -> 1 (but is_history_parent=False)
    graph = SessionGraph()
    add_node(graph, SessionNode(id=0, wait_after_ready=0.0))
    add_node(graph, SessionNode(id=1, wait_after_ready=0.0))
    add_edge(graph, SessionEdge(src=0, dst=1, is_history_parent=False))
    
    requests = {
        0: make_request(300),
        1: make_request(301),
    }
    session = Session(id=3, session_graph=graph, requests=requests)
    scheduler.schedule_session(session)
    
    # Complete 0
    req0 = pop_ready_with_timeout(scheduler)
    assert req0 is not None

    response_0 = ChannelResponse(
        modality=ChannelModality.TEXT,
        content="response_0"
    )
    scheduler.notify_completion(
        request_id=req0.id, 
        completed_at_monotonic=time.monotonic(), 
        success=True, 
        channel_responses={ChannelModality.TEXT: response_0}
    )
    
    # 1 ready, but no history inherited (empty list)
    req1 = pop_ready_with_timeout(scheduler)
    assert req1 is not None
    assert req1.history == []



import json
import os
from importlib.resources import files as _pkg_files
from typing import Dict, List

from veeksha.core.response import ChannelResponse


def store_generated_texts(
    output_dir: str, generated_responses: List[ChannelResponse]
) -> None:
    """Store generated responses in a text file."""
    with open(os.path.join(output_dir, "generated_texts.txt"), "w") as f:
        f.write(
            ("\n" + "-" * 30 + "\n").join([str(i.content) for i in generated_responses])
        )


def store_lmeval_results(output_dir: str, lmeval_results: Dict) -> None:
    """Store LMEval results in a JSON file."""
    with open(os.path.join(output_dir, "lmeval_results.json"), "w") as f:
        json.dump(lmeval_results, f, indent=4)


def load_corpus() -> List[str]:
    """Load corpus lines from packaged corpus.txt file"""
    corpus_resource = _pkg_files("veeksha.data").joinpath("corpus.txt")
    with corpus_resource.open("r", encoding="utf-8") as f:
        return f.readlines()



"""Utilities used by the benchmark runner."""

import hashlib
import os
import shutil
import time
from datetime import datetime
from typing import Any, Dict, Set, Tuple

import yaml
from tqdm import tqdm

from veeksha.config.benchmark import BenchmarkConfig
from veeksha.config.utils import dataclass_to_dict
from veeksha.core.seeding import SeedManager
from veeksha.evaluator.base import BaseEvaluator
from veeksha.evaluator.composite import CompositeEvaluator
from veeksha.evaluator.registry import EvaluatorRegistry
from veeksha.logger import init_logger
from veeksha.types import EvaluationType

logger = init_logger(__name__)

__all__ = [
    "_persist_config_yaml",
    "_init_output_dir",
    "build_evaluator",
    "maybe_run_warmup",
    "_monitor_for_completion",
]


def _persist_config_yaml(benchmark_config: BenchmarkConfig) -> str:
    """Write the resolved benchmark configuration to config.yml.

    Args:
        benchmark_config: The fully resolved benchmark configuration.

    Returns:
        Path to the persisted YAML file.
    """
    os.makedirs(benchmark_config.output_dir, exist_ok=True)
    config_dict = dataclass_to_dict(benchmark_config)
    config_path = os.path.join(benchmark_config.output_dir, "config.yml")
    with open(config_path, "w", encoding="utf-8") as config_file:
        yaml.safe_dump(
            config_dict,
            config_file,
            sort_keys=False,
            default_flow_style=False,
            allow_unicode=True,
        )
    logger.debug("Persisted benchmark config to %s", config_path)
    return config_path


def _init_output_dir(benchmark_config: BenchmarkConfig) -> str:
    """Resolve and prepare the final benchmark output directory.

    The function persists the config, computes its hash, and
    moves the config into a dated/hash-named subdirectory. The benchmark
    configuration's ``output_dir`` field is updated in-place to point to the
    resolved directory.

    Args:
        benchmark_config: Benchmark configuration to mutate.

    Returns:
        Path to the resolved output directory.
    """

    base_output_dir = benchmark_config.output_dir
    os.makedirs(base_output_dir, exist_ok=True)

    config_path = _persist_config_yaml(benchmark_config)
    with open(config_path, "rb") as config_file:
        config_bytes = config_file.read()
    config_hash = hashlib.sha1(config_bytes).hexdigest()[:8]

    timestamp_prefix = datetime.utcnow().strftime("%d:%m:%Y-%H:%M:%S")
    base_dir_name = f"{timestamp_prefix}-{config_hash}"
    resolved_output_dir = os.path.join(base_output_dir, base_dir_name)

    suffix = 1
    while os.path.exists(resolved_output_dir):
        suffix += 1
        resolved_output_dir = os.path.join(base_output_dir, f"{base_dir_name}-{suffix}")

    os.makedirs(resolved_output_dir, exist_ok=True)
    shutil.move(config_path, os.path.join(resolved_output_dir, "config.yml"))
    object.__setattr__(benchmark_config, "output_dir", resolved_output_dir)
    logger.info("Benchmark outputs will be stored in %s", resolved_output_dir)

    return resolved_output_dir


def maybe_run_warmup(session_generator, client) -> None:
    """Maybe run warmup sessions synchronously before benchmark.

    A warmup only runs the first request of each session specified.
    """

    import asyncio

    async def warmup_one(session):
        """Execute first request of a warmup session."""
        first_request = list(session.requests.values())[0]
        await client.send_request(first_request, session.id, 1)

    async def run_all(warmup_sessions):
        for session in tqdm(warmup_sessions, desc="Warmup", unit="sess"):
            await warmup_one(session)

    if hasattr(session_generator, "get_warmup_sessions"):
        warmup_sessions = session_generator.get_warmup_sessions()
        if warmup_sessions:
            logger.info(f"Running warmup with {len(warmup_sessions)} sessions")
            asyncio.run(run_all(warmup_sessions))
            logger.info("Warmup completed")


def build_evaluator(
    benchmark_config: BenchmarkConfig,
    *,
    seed_manager: SeedManager,
    session_generator: Any,
    benchmark_start_time: float,
) -> BaseEvaluator:
    """Build an evaluator instance (or composite evaluator) for a benchmark run.

    Notes:
    - Performance evaluator(s) are ordered first so that `CompositeEvaluator` uses
      performance for progress/timeout behavior.
    - Accuracy evaluation requires access to the session generator to map
      request IDs back to lm-eval instances.

    Args:
        benchmark_config: Benchmark configuration.
        seed_manager: Seed manager for reproducibility.
        session_generator: Session generator used for this run.
        benchmark_start_time: Run start time (monotonic), passed to evaluators for
            time-normalization and artifact timestamps.

    Returns:
        A `BaseEvaluator` (single evaluator or `CompositeEvaluator`).
    """
    evaluator_configs = sorted(
        benchmark_config.evaluators,
        key=lambda cfg: 0 if cfg.get_type() == EvaluationType.PERFORMANCE else 1,
    )

    evaluator_instances: list[BaseEvaluator] = []
    for cfg in evaluator_configs:
        kwargs: Dict[str, Any] = {
            "config": cfg,
            "seed_manager": seed_manager,
            "output_dir": f"{benchmark_config.output_dir}/metrics",
            "benchmark_start_time": benchmark_start_time,
        }
        if cfg.get_type() == EvaluationType.ACCURACY_LMEVAL:
            kwargs["session_generator"] = session_generator
        evaluator_instances.append(EvaluatorRegistry.get(cfg.get_type(), **kwargs))

    if not evaluator_instances:
        raise ValueError("BenchmarkConfig.evaluators must be non-empty.")

    return (
        evaluator_instances[0]
        if len(evaluator_instances) == 1
        else CompositeEvaluator(evaluator_instances)
    )


def _init_pbar(max_sessions: int, benchmark_timeout: float) -> Tuple[Any, bool]:
    """Initialize progress bar based on benchmark mode."""
    if max_sessions > 0:
        pbar = tqdm(
            total=max_sessions,
            desc="Sessions",
            unit="sess",
            dynamic_ncols=True,
            bar_format="{desc}: {n}/{total} [{percentage:3.0f}%] | {rate_fmt} | Elapsed: {elapsed}",
        )
        return pbar, False

    pbar = tqdm(
        total=int(benchmark_timeout),
        desc="Benchmark",
        unit="s",
        dynamic_ncols=True,
        bar_format="{desc}: {elapsed}/{total} s [{percentage:3.0f}%] | Sessions: {postfix}",
    )
    pbar.set_postfix_str("0")
    return pbar, True


def _update_pbar(
    pbar,
    time_based_progress: bool,
    elapsed: float,
    total_done: int,
    state: Dict[str, int],
) -> None:
    """Update progress bar with current state."""
    if time_based_progress:
        elapsed_int = int(elapsed)
        if elapsed_int > state["last_time_update"]:
            pbar.update(elapsed_int - state["last_time_update"])
            state["last_time_update"] = elapsed_int
        if total_done > state["last_completed"]:
            pbar.set_postfix_str(str(total_done))
            state["last_completed"] = total_done
        return

    if total_done > state["last_completed"]:
        pbar.update(total_done - state["last_completed"])
        state["last_completed"] = total_done


def _monitor_for_completion(
    traffic_scheduler,
    evaluator,
    pool_manager,
    benchmark_start: float,
    benchmark_timeout: float,
    timeout_triggered: bool,
    pre_timeout_request_ids: Set[str],
    max_sessions: int,
    post_timeout_grace_seconds: int = -1,
) -> Set[str]:
    """Observe worker progress and exit once requests settle.

    Returns:
        Set of request IDs that were still in-flight when monitoring stopped.
    """
    pbar, time_based_progress = _init_pbar(max_sessions, benchmark_timeout)
    pbar_state = {"last_completed": 0, "last_time_update": 0}
    timeout_start: float = 0.0
    in_flight_remaining: Set[str] = set()

    try:
        while True:
            time.sleep(0.1)

            completed, errored, _ = evaluator.get_session_counts()
            total_done = completed + errored
            elapsed = time.monotonic() - benchmark_start

            _update_pbar(pbar, time_based_progress, elapsed, total_done, pbar_state)

            if (
                not timeout_triggered
                and benchmark_timeout > 0
                and elapsed >= benchmark_timeout
            ):
                timeout_triggered = True
                timeout_start = time.monotonic()
                pre_timeout_request_ids = evaluator.get_registered_request_ids()
                in_flight = traffic_scheduler.get_in_flight_request_ids()
                pending = pre_timeout_request_ids & in_flight
                logger.info(
                    f"Benchmark timeout after {elapsed:.1f}s. "
                    f"Captured {len(pre_timeout_request_ids)} registered requests, "
                    f"{len(pending)} still in-flight."
                )

            prefetch_threads = pool_manager.thread_pools.get("prefetch", [])
            all_prefetch_done = all(not t.is_alive() for t in prefetch_threads)

            if timeout_triggered:
                current_in_flight = traffic_scheduler.get_in_flight_request_ids()
                remaining = pre_timeout_request_ids & current_in_flight

                # Check grace period if configured
                grace_elapsed = time.monotonic() - timeout_start
                if (
                    post_timeout_grace_seconds >= 0
                    and grace_elapsed >= post_timeout_grace_seconds
                ):
                    logger.warning(
                        f"Grace period of {post_timeout_grace_seconds}s expired. "
                        f"Force-exiting with {len(remaining)} requests still in-flight."
                    )
                    # Only include completed requests in metrics
                    completed_requests = pre_timeout_request_ids - remaining
                    evaluator.set_included_requests(completed_requests)
                    in_flight_remaining = remaining
                    break

                if not remaining:
                    logger.info("All pre-timeout requests completed")
                    evaluator.set_included_requests(pre_timeout_request_ids)
                    in_flight_remaining = set()
                    break
            elif all_prefetch_done and not traffic_scheduler.has_pending_work():
                logger.info("All sessions completed")
                in_flight_remaining = set()
                break
    finally:
        pbar.close()

    return in_flight_remaining



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import pytest

pytest.importorskip("lm_eval")

from veeksha.generator.session.lmeval import LMEvalSessionGenerator
from veeksha.types import ChannelModality
from veeksha.types import LMEvalOutputType


@dataclass
class _FakeTokenizer:
    def encode(self, text: str) -> list[int]:
        # Deterministic, minimal tokenizer for unit tests (not model-accurate).
        return list(range(len(text.split())))


@dataclass
class _FakeInstance:
    request_type: str
    args: tuple[Any, ...]


def test_lmeval_loglikelihood_uses_completions_max_tokens_param() -> None:
    gen = LMEvalSessionGenerator.__new__(LMEvalSessionGenerator)
    gen.text_tokenizer = _FakeTokenizer()
    gen.ctxlen_tokens_by_request_id = {}

    inst = _FakeInstance(
        request_type=str(LMEvalOutputType.LOGLIKELIHOOD),
        args=("Question: X\nAnswer:", " Y"),
    )
    req = LMEvalSessionGenerator._build_request(gen, request_id=0, instance=inst)  # type: ignore[arg-type]

    sp = req.metadata["sampling_params"]
    assert sp["stream"] is False
    assert sp["echo"] is True
    assert sp["logprobs"] == 1
    assert sp["temperature"] == 0
    assert sp["max_tokens"] == 1
    assert "max_completion_tokens" not in sp

    assert req.metadata["api_mode"] == "completions"
    assert ChannelModality.TEXT in req.channels





from veeksha.core.lazy_loader import _LazyLoader
from veeksha.types import ClientType
from veeksha.types.base_registry import BaseRegistry


class ClientRegistry(BaseRegistry):
    @classmethod
    def get_key_from_str(cls, key_str: str) -> ClientType:
        return ClientType.from_str(key_str)  # type: ignore


ClientRegistry.register(
    ClientType.OPENAI_CHAT_COMPLETIONS,
    _LazyLoader(
        "veeksha.client.openai_chat",
        "OpenAIChatCompletionsClient",
    ),
)

ClientRegistry.register(
    ClientType.OPENAI_ROUTER,
    _LazyLoader(
        "veeksha.client.openai_router",
        "OpenAIRouterClient",
    ),
)

ClientRegistry.register(
    ClientType.OPENAI_COMPLETIONS,
    _LazyLoader(
        "veeksha.client.openai_completions",
        "OpenAICompletionsClient",
    ),
)



import unittest
from veeksha.slo.metrics import extract_metric_values

class TestMetrics(unittest.TestCase):
    def test_extract_metric_values(self):
        metrics = {
            "ttfc": [0.1, 0.2, 0.3],
            "tbc": [[0.01, 0.02], [0.03, 0.04]],
            "end_to_end_latency": [1.0, 2.0, 3.0],
            "tpot": [0.001, 0.002, 0.003],
            "other": [10, 20]
        }

        # Test ttfc
        self.assertEqual(extract_metric_values("ttfc", metrics), [0.1, 0.2, 0.3])
        
        # Test tbc flattening
        self.assertEqual(extract_metric_values("tbc", metrics), [0.01, 0.02, 0.03, 0.04])
        
        # Test e2e mapping
        self.assertEqual(extract_metric_values("e2e", metrics), [1.0, 2.0, 3.0])
        
        # Test tpot mapping
        self.assertEqual(extract_metric_values("tpot", metrics), [0.001, 0.002, 0.003])
        
        # Test unknown
        self.assertEqual(extract_metric_values("unknown", metrics), [])
        
        # Test empty
        self.assertEqual(extract_metric_values("ttfc", {}), [])

if __name__ == "__main__":
    unittest.main()
